import logging
import os
from typing import List, Union, Dict, Any, Optional, Set
from types import ModuleType
from collections import Counter, defaultdict

from databricks.feature_store.utils import validation_utils
from databricks.feature_store.constants import (
    PREDICTION_COLUMN_NAME,
    MODEL_DATA_PATH_ROOT,
    _WARN,
)
from databricks.feature_store.entities.feature_lookup import FeatureLookup
from databricks.feature_store.entities.feature import Feature
from databricks.feature_store.entities.feature_column_info import FeatureColumnInfo
from databricks.feature_store.entities.feature_table_info import FeatureTableInfo
from databricks.feature_store.entities.feature_spec import FeatureSpec
from databricks.feature_store.entities.feature_table import FeatureTable
from databricks.feature_store.entities.source_data_column_info import (
    SourceDataColumnInfo,
)
from databricks.feature_store.hive_client import HiveClient
from databricks.feature_store.catalog_client import CatalogClient
from databricks.feature_store import mlflow_model_constants
from databricks.feature_store.training_set import TrainingSet
from databricks.feature_store.utils import utils
from databricks.feature_store.utils.feature_lookup_utils import join_feature_data
from databricks.feature_store.utils import request_context
from databricks.feature_store.utils.request_context import (
    RequestContext,
)
from databricks.feature_store.version import VERSION
from databricks.feature_store.utils import schema_utils
from databricks.feature_store._hive_client_helper import HiveClientHelper
from databricks.feature_store._catalog_client_helper import CatalogClientHelper

from pyspark.sql import DataFrame
from pyspark.sql.functions import struct

from mlflow.models import Model
from mlflow.utils.file_utils import TempDir, read_yaml
import mlflow

_logger = logging.getLogger(__name__)


class TrainingScoringClient:
    def __init__(
        self,
        catalog_client: CatalogClient,
        catalog_client_helper: CatalogClientHelper,
        hive_client: HiveClient,
        model_registry_uri: str,
    ):
        self._catalog_client = catalog_client
        self._catalog_client_helper = catalog_client_helper
        self._hive_client = hive_client
        self._hive_client_helper = HiveClientHelper(self._hive_client)
        self._model_registry_uri = model_registry_uri

    def create_training_set(
        self,
        df: DataFrame,
        feature_lookups: List[FeatureLookup],
        label: Union[str, List[str], None],
        exclude_columns: List[str] = [],
    ) -> TrainingSet:

        req_context = RequestContext(request_context.CREATE_TRAINING_SET)

        validation_utils.check_dataframe_type(df)

        # Verify DataFrame does not have duplicate columns
        df_column_name_counts = Counter(df.columns)
        for column_name in df.columns:
            if df_column_name_counts[column_name] > 1:
                raise ValueError(
                    f"Found duplicate DataFrame column name '{column_name}'"
                )

        # Initialize label_names with empty list if label is not provided
        label_names = utils.as_list(label, [])

        # Verify that label is in DataFrame and not in exclude_columns
        for label_name in label_names:
            if label_name not in df.columns:
                raise ValueError(
                    f"Label column '{label_name}' was not found in DataFrame"
                )
            if label_name in exclude_columns:
                raise ValueError(f"Label column '{label_name}' cannot be excluded")

        table_names = set([fl.table_name for fl in feature_lookups])

        feature_table_features_map = self._get_feature_names_for_tables(
            req_context, table_names=table_names
        )
        feature_table_metadata_map = self._get_feature_table_metadata_for_tables(
            req_context, table_names=table_names
        )
        feature_table_data_map = self._load_feature_data_for_tables(
            table_names=table_names
        )

        # Collect FeatureColumnInfos
        feature_column_infos = self._explode_feature_lookups(
            feature_lookups, feature_table_features_map, feature_table_metadata_map
        )

        feature_output_names = [
            feature_info.output_name for feature_info in feature_column_infos
        ]

        # Verify features have unique output names
        feature_output_names_counts = Counter(feature_output_names)
        for output_name in feature_output_names_counts:
            if feature_output_names_counts[output_name] > 1:
                raise ValueError(f"Found duplicate feature output name '{output_name}'")

        # Verify labels do not collide with feature output names
        for label_name in label_names:
            if label_name in feature_output_names:
                raise ValueError(
                    f"Feature cannot have same output name as label '{label_name}'"
                )

        # Verify that DataFrame does not override features requested in FeatureLookup list
        overridden_features = [
            column_name
            for column_name in df.columns
            if column_name in feature_output_names
        ]
        if len(overridden_features) > 0:
            overriden_features_names = ", ".join(overridden_features)
            raise ValueError(
                f"DataFrame contains column names that match feature output names specified"
                f" in FeatureLookups '{overriden_features_names}'. Either remove these columns"
                f" from the DataFrame or FeatureLookups."
            )

        # Collect SourceDataColumnInfo
        source_columns_infos = [
            SourceDataColumnInfo(column_name)
            for column_name in df.columns
            if column_name not in feature_output_names
            and column_name not in label_names
        ]

        column_infos = [
            ci
            for ci in (source_columns_infos + feature_column_infos)
            if ci.output_name not in exclude_columns
        ]

        # Keep table_infos sorted by table name, so they appear sorted in feature_spec.yaml
        table_infos = sorted(
            [
                FeatureTableInfo(
                    table_name=feature_table_name,
                    table_id=feature_table_metadata.table_id,
                )
                for feature_table_name, feature_table_metadata in feature_table_metadata_map.items()
            ],
            key=lambda table_info: table_info.table_name,
        )

        workspace_id = self._catalog_client.feature_store_workspace_id

        # Build FeatureSpec
        feature_spec = FeatureSpec(column_infos, table_infos, workspace_id, VERSION)

        # Validate FeatureSpec
        self._validate_data_for_feature_spec(
            feature_spec.feature_column_infos,
            feature_table_features_map,
            feature_table_data_map,
        )

        # TODO(divyagupta-db): Move validation from _validate_join_feature_data in feature_lookup_utils.py
        #  to a helper function called here and in score_batch.

        # Add consumer of each feature as final step
        consumer_feature_table_map = defaultdict(list)
        for feature in feature_column_infos:
            consumer_feature_table_map[feature.table_name].append(feature.feature_name)
        self._catalog_client_helper.add_consumers(
            consumer_feature_table_map, req_context
        )

        return TrainingSet(
            feature_spec,
            df,
            label_names,
            feature_table_metadata_map,
            feature_table_data_map,
        )

    def score_batch(
        self, model_uri: str, df: DataFrame, result_type: str = "double"
    ) -> DataFrame:
        req_context = RequestContext(request_context.SCORE_BATCH)

        validation_utils.check_dataframe_type(df)

        if PREDICTION_COLUMN_NAME in df.columns:
            raise ValueError(
                "FeatureStoreClient.score_batch returns a DataFrame with a new column "
                f'"{PREDICTION_COLUMN_NAME}". df already has a column with name '
                f'"{PREDICTION_COLUMN_NAME}".'
            )

        # The call to mlflow.set_registry_uri sets the model registry globally, so it needs
        # to be set appropriately each time. If no model_registry_uri is specified, we pass
        # in an empty string to reset the model registry to the local one.
        mlflow.set_registry_uri(self._model_registry_uri or "")

        artifact_path = os.path.join(mlflow.pyfunc.DATA, MODEL_DATA_PATH_ROOT)

        with TempDir() as tmp_location:
            local_path = utils.download_model_artifacts(model_uri, tmp_location.path())
            model_data_path = os.path.join(local_path, artifact_path)
            feature_spec = FeatureSpec.load(model_data_path)
            raw_model_path = os.path.join(
                model_data_path, mlflow_model_constants.RAW_MODEL_FOLDER
            )
            predict_udf = self._hive_client.get_predict_udf(
                raw_model_path, result_type=result_type
            )
            # TODO (ML-17260) Consider reading the timestamp from the backend instead of feature store artifacts
            ml_model = Model.load(
                os.path.join(local_path, mlflow_model_constants.ML_MODEL)
            )
            model_creation_timestamp_ms = (
                utils.utc_timestamp_ms_from_iso_datetime_string(
                    ml_model.utc_time_created
                )
            )

        # Validate that columns needed for joining feature tables exist and are not duplicates.
        required_cols = []
        for column_info in feature_spec.column_infos:
            col_info_req_cols = (
                [column_info.name]
                if isinstance(column_info, SourceDataColumnInfo)
                else column_info.lookup_key
            )
            for col in col_info_req_cols:
                if col not in required_cols:
                    required_cols.append(col)

        df_column_name_counts = Counter(df.columns)
        for col in required_cols:
            if col not in df.columns:
                raise ValueError(
                    f"DataFrame is missing column '{col}'. The following columns are required: "
                    f"{required_cols}."
                )
            if df_column_name_counts[col] > 1:
                raise ValueError(
                    f"Columns required for scoring or looking up features cannot be "
                    f"duplicates. Found duplicate column '{col}'."
                )

        table_names = set([fci.table_name for fci in feature_spec.feature_column_infos])

        feature_table_features_map = self._get_feature_names_for_tables(
            req_context, table_names=table_names
        )
        feature_table_metadata_map = self._get_feature_table_metadata_for_tables(
            req_context, table_names=table_names
        )
        feature_table_data_map = self._load_feature_data_for_tables(
            table_names=table_names
        )

        self._validate_data_for_feature_spec(
            feature_spec.feature_column_infos,
            feature_table_features_map,
            feature_table_data_map,
        )

        # Check if the fetched feature tables match the feature tables logged in training

        # 1. Compare feature table ids
        # Check for feature_spec logged with client versions that supports table_infos
        if len(feature_spec.table_infos) > 0:
            # When feature_spec.yaml is parsed, FeatureSpec.load will assert
            # that the listed table names in input_tables match table names in input_columns.
            # The following code assumes this as invariant and only checks for the table IDs.
            mismatched_tables = []
            for table_info in feature_spec.table_infos:
                feature_table = feature_table_metadata_map[table_info.table_name]
                if feature_table and table_info.table_id != feature_table.table_id:
                    mismatched_tables.append(table_info.table_name)
            if len(mismatched_tables) > 0:
                plural = len(mismatched_tables) > 1
                _logger.warning(
                    f"Feature table{'s' if plural else ''} {', '.join(mismatched_tables)} "
                    f"{'were' if plural else 'was'} deleted and recreated after "
                    f"the model was trained. Model performance may be affected if the features "
                    f"used in scoring have drifted from the features used in training."
                )

        # 2. Compare model creation timestamp with feature table creation timestamps
        feature_tables_created_after_model = []
        for name, metadata in feature_table_metadata_map.items():
            if model_creation_timestamp_ms < metadata.creation_timestamp:
                feature_tables_created_after_model.append(name)
        if len(feature_tables_created_after_model) > 0:
            plural = len(feature_tables_created_after_model) > 1
            message = (
                f"Feature table{'s' if plural else ''} {', '.join(feature_tables_created_after_model)} "
                f"{'were' if plural else 'was'} created after the model was logged. "
                f"Model performance may be affected if the features used in scoring have drifted "
                f"from the features used in training."
            )
            _logger.warning(message)

        df_with_features = join_feature_data(
            feature_spec,
            df,
            extra_columns=df.columns,
            feature_table_metadata_map=feature_table_metadata_map,
            feature_table_data_map=feature_table_data_map,
        )

        udf_input_columns = [c.output_name for c in feature_spec.column_infos]
        udf_inputs = struct(*udf_input_columns)

        df_with_predictions = df_with_features.withColumn(
            PREDICTION_COLUMN_NAME, predict_udf(udf_inputs)
        )

        # Reorder dataframe columns
        df_columns = [col for col in df.columns]
        extra_feature_spec_columns = [
            col for col in udf_input_columns if col not in df_columns
        ]
        prediction_column = [PREDICTION_COLUMN_NAME]

        return_value = df_with_predictions.select(
            df_columns + extra_feature_spec_columns + prediction_column
        )

        # Add consumer of each feature as final step
        consumer_feature_table_map = defaultdict(list)
        for feature in feature_spec.feature_column_infos:
            consumer_feature_table_map[feature.table_name].append(feature.feature_name)
        self._catalog_client_helper.add_consumers(
            consumer_feature_table_map, req_context
        )

        return return_value

    def log_model(
        self,
        model: Any,
        artifact_path: str,
        *,
        flavor: ModuleType,
        training_set: TrainingSet,
        registered_model_name: Optional[str] = None,
        await_registration_for: int = mlflow.tracking._model_registry.DEFAULT_AWAIT_MAX_SLEEP_SECONDS,
        **kwargs,
    ):
        with TempDir() as tmp_location:
            data_path = os.path.join(tmp_location.path(), "feature_store")
            raw_mlflow_model = Model()
            raw_model_path = os.path.join(
                data_path, mlflow_model_constants.RAW_MODEL_FOLDER
            )
            if flavor.FLAVOR_NAME != mlflow.pyfunc.FLAVOR_NAME:
                flavor.save_model(
                    model, raw_model_path, mlflow_model=raw_mlflow_model, **kwargs
                )
            else:
                flavor.save_model(
                    raw_model_path,
                    mlflow_model=raw_mlflow_model,
                    python_model=model,
                    **kwargs,
                )
            if not "python_function" in raw_mlflow_model.flavors:
                raise ValueError(
                    f"FeatureStoreClient.log_model does not support '{flavor.__name__}' "
                    f"since it does not have a python_function model flavor."
                )

            # Re-use the conda environment from the raw model for the packaged model. Later, we may
            # add an additional requirement for the Feature Store library. At the moment, however,
            # the databricks-feature-store package is not available via conda or pip.
            conda_file = raw_mlflow_model.flavors["python_function"][mlflow.pyfunc.ENV]
            conda_env = read_yaml(raw_model_path, conda_file)

            # Get the pip package string for the databricks-feature-lookup client
            databricks_feature_lookup_pip_package = utils.pip_depependency_pinned_major_version(
                pip_package_name=mlflow_model_constants.FEATURE_LOOKUP_CLIENT_PIP_PACKAGE,
                major_version=mlflow_model_constants.FEATURE_LOOKUP_CLIENT_MAJOR_VERSION,
            )

            # Add pip dependencies required for online feature lookups
            utils.add_mlflow_pip_depependency(
                conda_env, databricks_feature_lookup_pip_package
            )

            training_set.feature_spec.save(data_path)

            # Log the packaged model. If no run is active, this call will create an active run.
            mlflow.pyfunc.log_model(
                artifact_path=artifact_path,
                loader_module=mlflow_model_constants.MLFLOW_MODEL_NAME,
                data_path=data_path,
                code_path=None,
                conda_env=conda_env,
            )
        if registered_model_name is not None:
            # The call to mlflow.pyfunc.log_model will create an active run, so it is safe to
            # obtain the run_id for the active run.
            run_id = mlflow.tracking.fluent.active_run().info.run_id

            # The call to mlflow.set_registry_uri sets the model registry globally, so it needs
            # to be set appropriately each time. If no model_registry_uri is specified, we pass
            # in an empty string to reset the model registry to the local one.
            mlflow.set_registry_uri(self._model_registry_uri or "")
            mlflow.register_model(
                "runs:/%s/%s" % (run_id, artifact_path),
                registered_model_name,
                await_registration_for=await_registration_for,
            )

    def _get_feature_names_for_tables(
        self, req_context: RequestContext, table_names: Set[str]
    ) -> Dict[str, List[Feature]]:
        """
        Lookup features from the feature catalog for all table_names, return a dictionary of tablename -> list of features.
        """
        return {
            table_name: self._catalog_client.get_features(table_name, req_context)
            for table_name in table_names
        }

    def _get_feature_table_metadata_for_tables(
        self, req_context: RequestContext, table_names: Set[str]
    ) -> Dict[str, FeatureTable]:
        """
        Lookup FeatureTable metadata from the feature catalog for all table_names, return a dictionary of tablename -> FeatureTable.
        """
        return {
            table_name: self._catalog_client.get_feature_table(table_name, req_context)
            for table_name in table_names
        }

    def _load_feature_data_for_tables(
        self, table_names: Set[str]
    ) -> Dict[str, DataFrame]:
        """
        Load feature DataFrame objects for all table_names, return a dictionary of tablename -> DataFrame.
        """
        return {
            table_name: self._hive_client.read_table(table_name)
            for table_name in table_names
        }

    def _explode_feature_lookups(
        self,
        feature_lookups: List[FeatureLookup],
        feature_table_features_map: Dict[str, List[Feature]],
        feature_table_metadata_map: Dict[str, FeatureTable],
    ) -> List[FeatureColumnInfo]:
        """
        Explode FeatureLookups and collect into FeatureColumnInfos.  A FeatureLookup may explode into either:

        1. A single FeatureColumnInfo, in the case where only a single feature name is specified.
        2. Multiple FeatureColumnInfos, in the cases where either multiple or all feature names are specified.

        Additionally, when all feature names are specified in a FeatureLookup via setting feature_names to None,
        FeatureColumnInfos will be created for all features except primary keys.
        The order of the FeatureColumnInfos returned by this method will be the same order as returned by
        the backend:

        - All partition keys that are not primary keys, in the partition key order
        - All other non-key features
        """
        feature_column_infos = []
        for feature_lookup in feature_lookups:
            feature_column_infos_for_feature_lookup = self._explode_feature_lookup(
                feature_lookup=feature_lookup,
                features=feature_table_features_map[feature_lookup.table_name],
                feature_table=feature_table_metadata_map[feature_lookup.table_name],
            )
            feature_column_infos += feature_column_infos_for_feature_lookup
        return feature_column_infos

    def _explode_feature_lookup(
        self,
        feature_lookup: FeatureLookup,
        features: List[Feature],
        feature_table: FeatureTable,
    ) -> List[FeatureColumnInfo]:
        feature_names = []
        if feature_lookup._get_feature_names():
            # If the user explicitly passed in a feature name or list of feature names, use that
            feature_names += feature_lookup._get_feature_names()
        else:
            # Otherwise assume the user wants all columns in the feature table
            feature_names += [
                feature.name
                for feature in features
                # Filter out primary keys and timestamp keys
                if (
                    feature.name
                    not in [*feature_table.primary_keys, *feature_table.timestamp_keys]
                )
            ]

        return [
            FeatureColumnInfo(
                table_name=feature_lookup.table_name,
                feature_name=feature_name,
                lookup_key=utils.as_list(feature_lookup.lookup_key),
                output_name=(feature_lookup._get_output_name(feature_name)),
                timestamp_lookup_key=utils.as_list(
                    feature_lookup.timestamp_lookup_key, default=[]
                ),
            )
            for feature_name in feature_names
        ]

    def _validate_data_for_feature_spec(
        self,
        feature_column_infos: List[FeatureColumnInfo],
        features_by_table: Dict[str, List[Feature]],
        feature_table_data_map: Dict[str, DataFrame],
    ):
        """
        Validate that

        1) Feature tables exist in Delta
        2) Features exist in Delta and Feature Catalog
        3) Feature types match in Delta and Feature Catalog
        """

        table_names = set([fi.table_name for fi in feature_column_infos])

        # Validate that all feature tables exist in Hive to fail-fast before API requests
        for table_name in table_names:
            self._hive_client_helper.check_feature_table_exists_in_hive(table_name)

        table_data = {}
        for table_name in table_names:
            feature_names_in_spec = [
                info.feature_name
                for info in feature_column_infos
                if info.table_name == table_name
            ]

            catalog_features = features_by_table[table_name]

            feature_table_data = feature_table_data_map[table_name]

            catalog_schema = {
                feature.name: feature.data_type for feature in catalog_features
            }
            delta_schema = {
                feature.name: feature.dataType.typeName().upper()
                for feature in feature_table_data.schema
            }

            for feature_name in feature_names_in_spec:
                if feature_name not in catalog_schema:
                    raise ValueError(
                        f"Unable to find feature '{feature_name}' from feature table '{table_name}' in Feature Catalog."
                    )
                if feature_name not in delta_schema:
                    raise ValueError(
                        f"Unable to find feature '{feature_name}' from feature table '{table_name}' in Delta."
                    )
                if catalog_schema[feature_name] != delta_schema[feature_name]:
                    raise ValueError(
                        f"Expected type of feature '{feature_name}' from feature table '{table_name}' "
                        f"to be equivalent in Feature Catalog and Delta. "
                        f"Feature has type '{catalog_schema[feature_name]}' in Feature Catalog and "
                        f"'{delta_schema[feature_name]}' in Delta."
                    )

            # Warn if mismatch in other features in feature table
            if not schema_utils.catalog_matches_delta_schema(
                catalog_features, feature_table_data.schema
            ):
                schema_utils.log_catalog_schema_not_match_delta_schema(
                    catalog_features,
                    feature_table_data.schema,
                    level=_WARN,
                )



"""Module for configure and compute the data profiles for feature tables."""

import functools
from typing import Iterable, List, Optional

from pyspark.sql import Column, DataFrame, functions as F, types as T
from databricks.feature_store.entities.data_profiles_entities import (
    Granularities,
    ComputationMode,
)
from databricks.feature_store import FeatureStoreClient

_GRANULARITY_COLUMN = "granularity"
_WINDOW_COLUMN = "window"


class DataProfilesSpec:
    def __init__(
        self,
        granularities: Iterable[str] = ("GLOBAL",),
        computation_mode: str = "MANUAL",
        features: Optional[Iterable[str]] = None,
    ):
        """
        This class serves as a configuration of the data profiles.

        :param granularities: Defines the window size when aggregating data based on the table's `timestamp_keys`.
          Currently only the following granularities are supported: {“GLOBAL“, “1 HOUR”, “1 DAY”, “1 WEEK”, “1 MONTH“}.
          “GLOBAL“ means the whole table will be grouped together.
        :param computation_mode: controls when a profile is re-computed. Valid values:
          - "MANUAL": No automatic triggering, user has to invoke manually by calling `refresh_data_profiles` or
                      through the UI.
          - "AUTO": Profiles are re-computed automatically when the feature table is updated. Streaming update is not
                    supported on this mode.
        :param features: A list of table column names to be computed for the data profiles. If this field is None,
          then all the columns will be included.
        """
        supported_computation_modes = [c.value for c in ComputationMode]
        if computation_mode.upper() not in supported_computation_modes:
            raise ValueError(
                f"Unsupported computation_mode: '{computation_mode}'. Use one of {', '.join(supported_computation_modes)}."
            )

        supported_granularities = [g.value for g in Granularities]
        for granularity in granularities:
            if granularity.upper() not in supported_granularities:
                raise ValueError(
                    f"Unsupported granularities: '{granularity}'. Use one of {', '.join(supported_granularities)}."
                )

        self._granularities = [
            Granularities(granularity.upper()) for granularity in granularities
        ]
        self._computation_mode = ComputationMode(computation_mode.upper())
        self._features = features

    @property
    def granularities(self) -> Iterable[str]:
        return [granularity.value for granularity in self._granularities]

    @property
    def computation_mode(self) -> str:
        return self._computation_mode.value

    @property
    def features(self):
        return self._features


def _generate_window_col_exp(timestamp_col: str, granularity: str) -> Column:
    """
    Generates the window column that will be used for aggregation.

    :param timestamp_col: The name of the timestamp column
    :param granularity: The granularity of the time window grouping
    :return: A struct columns with the nested fields 'start' and 'end'
    """
    if granularity == Granularities.GLOBAL:
        # Use NULL to represent window start/end for global aggregation
        return F.struct(F.lit(None).alias("start"), F.lit(None).alias("end"))
    elif granularity == Granularities.ONE_MONTH:
        # "1 month" is not explicitly supported by F.window, so we need to construct
        # the output manually with a struct.
        start_date = F.date_trunc("month", F.col(timestamp_col)).alias("start")
        end_date = F.date_trunc("month", F.add_months(start_date, 1)).alias("end")
        return F.struct(start_date, end_date)
    else:
        # If the granularity is 1 week, then we need to shift the window calculation by 4 days,
        # since the start of the epoch falls on Th.
        # This shift makes week windows start on Monday and end on Sunday.
        start_time = "4 days" if granularity == Granularities.ONE_WEEK else None
        # Note that this function assumes a compatible granularity string, see
        # https://spark.apache.org/docs/latest/api/python/reference/api/pyspark.sql.functions.window.html
        # Spark natively supports "1 hour", "1 day", and "1 week".
        return F.window(
            F.col(timestamp_col),
            windowDuration=granularity,
            slideDuration=granularity,
            startTime=start_time,
        )


def _compute_data_profiles(
    df: DataFrame, spec: DataProfilesSpec, ts_keys: List[str]
) -> DataFrame:
    """
    This internal function computes the data profiles given the DataFrame.

    Uses the provided timestamp keys and profile spec to compute a profile
    DataFrame for every granularity, and unions the resulting DataFrames
    into a single output.

    :param df: The data to compute the profiles.
    :param spec: The configuration that defines how profiles should be computed.
    :param ts_keys: The timestamp keys associated with the table.
    :return: The computed data profiles in a DataFrame with columns:
                - granularity
                - window
                - column_name
                - <all computed statistics>
    """
    if not ts_keys and spec.granularities != ["GLOBAL"]:
        raise ValueError(
            "Only 'GLOBAL' granularity is supported for tables without timestamp keys."
        )

    # Feature tables can have at most one timestamp key
    ts_key = ts_keys[0] if ts_keys else None

    features_to_compute = spec.features if spec.features else df.columns

    # TODO(ML-21787): Refactor so that we can reuse the metrics definition here.
    # Generates a struct that contains an expression for every aggregate metric we want to compute.
    aggregates_expressions = [
        F.struct(
            [
                F.count(column_name).alias("count"),
                F.sum(F.col(column_name).isNull().cast(T.LongType())).alias(
                    "num_nulls"
                ),
            ]
        ).alias(column_name)
        for column_name in features_to_compute
    ]
    # For each feature, we map the column name to itself and explode it so we can compute
    # the same set of aggregate expressions on every input feature
    map_args_for_explode = []
    for column_name in features_to_compute:
        map_args_for_explode.extend([F.lit(column_name), F.col(column_name)])
    transposed_columns = F.explode(F.create_map(*map_args_for_explode))

    # We will compute an aggregate DataFrame for every granularity from the spec and union them at the end
    aggregate_dfs = []
    for granularity in spec.granularities:
        # Create the granularity and window column to define this aggregation grouping,
        # and preserve these meta columns in the output schema.
        granularity_col = F.lit(granularity).alias(_GRANULARITY_COLUMN)
        window_col = _generate_window_col_exp(ts_key, granularity).alias(_WINDOW_COLUMN)
        meta_columns = [granularity_col, window_col]
        meta_column_names = [F.col(_GRANULARITY_COLUMN), F.col(_WINDOW_COLUMN)]

        aggregate_dfs.append(
            df.groupBy(meta_columns)
            .agg(*aggregates_expressions)
            .select(meta_column_names + [transposed_columns])
            .select(
                meta_column_names
                + [
                    F.col("key").alias("column_name"),
                    F.col("value.count"),
                    F.col("value.num_nulls"),
                ]
            )
        )

    return functools.reduce(DataFrame.unionByName, aggregate_dfs)


def configure_data_profiles(table_name: str, data_profiles_spec: DataProfilesSpec):
    """
    Overwrites the configuration of the data profiles of the feature table. Only those with `Can Manage` permission
    on the table can call this function.

    :param table_name: A feature table name of the form <database_name>.<table_name>, for example dev.user_features.
    :param data_profiles_spec: An instance of `DataProfilesSpec` which defined the behavior of the data profiles
      computation.
    """
    raise NotImplementedError()


def compute_data_profiles(
    table_name: str,
    granularities: Iterable[str] = ("GLOBAL",),
    features: Optional[Iterable[str]] = None,
):
    """
    Compute and return data profiles of the table. This will not overwrite the data profiles saved on the feature
    table. Only those with `Can View Metadata` permission on the table can call this function.

    :param table_name: A feature table name of the form <database_name>.<table_name>, for example dev.user_features.
    :param granularities: List of granularities to use when aggregating data into time windows based on
      the table's `timestamp_keys`.
      Currently only the following granularities are supported: {“GLOBAL“, “1 HOUR”, “1 DAY”, “1 WEEK”, “1 MONTH“}.
      If None, then the whole table will be grouped together.
    :param features: A list of table column names to be computed for the data profiles. If this field is None,
          then all the columns will be included.
    """
    fs = FeatureStoreClient()
    df = fs.read_table(table_name)
    ts_keys = fs.get_table(table_name).timestamp_keys
    spec = DataProfilesSpec(granularities=granularities, features=features)
    return _compute_data_profiles(df, spec, ts_keys)


def refresh_data_profiles(table_name: str):
    """
    Refresh the data profiles of the table. It re-computes the data profiles and saves the result. The data profiles
    must have been enabled before calling this API. Only the one with `Can Edit Metadata` permission on the table can
    call this function.

    :param table_name: A feature table name of the form <database_name>.<table_name>, for example dev.user_features.
    """
    raise NotImplementedError()



from databricks.feature_store.api.proto.feature_catalog_pb2 import (
    PermissionLevel as ProtoPermissionLevel,
)
from databricks.feature_store.entities._proto_enum_entity import _ProtoEnumEntity

from typing import Any


class _PermissionLevel(_ProtoEnumEntity):
    """Permission Levels."""

    _CAN_MANAGE = ProtoPermissionLevel.Value("CAN_MANAGE")
    _CAN_EDIT_METADATA = ProtoPermissionLevel.Value("CAN_EDIT_METADATA")
    _CAN_VIEW_METADATA = ProtoPermissionLevel.Value("CAN_VIEW_METADATA")

    @classmethod
    def _enum_type(cls) -> Any:
        return ProtoPermissionLevel

    @staticmethod
    def can_write_to_catalog(permission_level):
        return permission_level in [
            _PermissionLevel._CAN_MANAGE,
            _PermissionLevel._CAN_EDIT_METADATA,
        ]



import base64
import time
import logging
import json

import requests

from databricks.feature_store.version import VERSION
from mlflow.protos import databricks_pb2
from google.protobuf.json_format import MessageToJson, ParseDict
from databricks.feature_store.utils.request_context import RequestContext

_REST_API_PATH_PREFIX = "/api/2.0"
_DEFAULT_HEADERS = {"User-Agent": "feature-store-python-client/%s" % VERSION}

_logger = logging.getLogger(__name__)


def http_request(
    host_creds,
    endpoint,
    retries=3,
    retry_interval=3,
    max_rate_limit_interval=60,
    extra_headers=None,
    **kwargs
):
    """
    Makes an HTTP request with the specified method to the specified hostname/endpoint. Ratelimit
    error code (429) will be retried with an exponential back off (1, 2, 4, ... seconds) for at most
    `max_rate_limit_interval` seconds.  Internal errors (500s) will be retried up to `retries` times
    , waiting `retry_interval` seconds between successive retries. Parses the API response
    (assumed to be JSON) into a Python object and returns it.

    :param extra_headers: a dictionary of extra headers to add to request.
    :param host_creds: Databricks creds containing hostname and optional authentication.
    :return: Parsed API response
    """
    hostname = host_creds.host
    auth_str = None
    if host_creds.username and host_creds.password:
        basic_auth_str = ("%s:%s" % (host_creds.username, host_creds.password)).encode(
            "utf-8"
        )
        auth_str = "Basic " + base64.standard_b64encode(basic_auth_str).decode("utf-8")
    elif host_creds.token:
        auth_str = "Bearer %s" % host_creds.token

    headers = dict(_DEFAULT_HEADERS)

    # Inject any extra headers
    if extra_headers is not None:
        headers.update(extra_headers)

    if auth_str:
        headers["Authorization"] = auth_str

    if host_creds.server_cert_path is None:
        verify = not host_creds.ignore_tls_verification
    else:
        verify = host_creds.server_cert_path

    if host_creds.client_cert_path is not None:
        kwargs["cert"] = host_creds.client_cert_path

    def request_with_ratelimit_retries(max_rate_limit_interval, **kwargs):
        response = requests.request(**kwargs)
        time_left = max_rate_limit_interval
        sleep = 1
        while response.status_code == 429 and time_left > 0:
            _logger.warning(
                "API request to %s returned status code 429 (Rate limit exceeded). "
                "Retrying in %d seconds. "
                "Will continue to retry 429s for up to %d seconds.",
                kwargs.get("url", endpoint),
                sleep,
                time_left,
            )
            time.sleep(sleep)
            time_left -= sleep
            response = requests.request(**kwargs)
            sleep = min(time_left, sleep * 2)  # sleep for 1, 2, 4, ... seconds;
        return response

    cleaned_hostname = hostname[:-1] if hostname.endswith("/") else hostname
    url = "%s%s" % (cleaned_hostname, endpoint)
    for i in range(retries):
        try:
            response = request_with_ratelimit_retries(
                max_rate_limit_interval,
                url=url,
                headers=headers,
                verify=verify,
                **kwargs
            )
            if response.status_code < 500:
                return response
            else:
                _logger.error(
                    "API request to %s failed with code %s, retrying up to %s more times. "
                    "API response body: %s",
                    url,
                    response.status_code,
                    retries - i - 1,
                    response.text,
                )
        # All exceptions that Requests explicitly raises inherit from requests.exceptions.RequestException.
        # See https://docs.python-requests.org/en/latest/user/quickstart/#errors-and-exceptions
        # for more details.
        except requests.exceptions.RequestException as e:
            _logger.error(
                "API request encountered unexpected error: %s. "
                "Requested service might be temporarily unavailable, "
                "retrying up to %s more times.",
                str(e),
                retries - i - 1,
            )
        time.sleep(retry_interval)
    raise Exception("API request to %s failed after %s tries" % (url, retries))


def _can_parse_as_json(string):
    try:
        json.loads(string)
        return True
    except Exception:  # pylint: disable=broad-except
        return False


def verify_rest_response(response, endpoint):
    """Verify the return code and format, raise exception if the request was not successful."""
    if response.status_code != 200:
        if _can_parse_as_json(response.text):
            # ToDo(ML-20622): return cleaner error to client, eg: mlflow.exceptions.RestException
            raise Exception(json.loads(response.text))
        else:
            base_msg = (
                "API request to endpoint %s failed with error code "
                "%s != 200"
                % (
                    endpoint,
                    response.status_code,
                )
            )
            raise Exception("%s. Response body: '%s'" % (base_msg, response.text))

    # Skip validation for endpoints (e.g. DBFS file-download API) which may return a non-JSON
    # response
    if endpoint.startswith(_REST_API_PATH_PREFIX) and not _can_parse_as_json(
        response.text
    ):
        base_msg = (
            "API request to endpoint was successful but the response body was not "
            "in a valid JSON format"
        )
        raise Exception("%s. Response body: '%s'" % (base_msg, response.text))

    return response


def get_error_code(e: Exception):
    if hasattr(e, "args") and len(e.args) > 0 and "error_code" in e.args[0]:
        return e.args[0]["error_code"]
    return None


def get_path(path_prefix, endpoint_path):
    return "{}{}".format(path_prefix, endpoint_path)


def extract_api_info_for_service(service, path_prefix):
    """Return a dictionary mapping each API method to a tuple (path, HTTP method)"""
    service_methods = service.DESCRIPTOR.methods
    res = {}
    for service_method in service_methods:
        endpoints = service_method.GetOptions().Extensions[databricks_pb2.rpc].endpoints
        endpoint = endpoints[0]
        endpoint_path = get_path(path_prefix, endpoint.path)
        res[service().GetRequestClass(service_method)] = (
            endpoint_path,
            endpoint.method,
        )
    return res


def json_to_proto(js_dict, message):
    """Parses a JSON dictionary into a message proto, ignoring unknown fields in the JSON."""
    ParseDict(js_dict=js_dict, message=message, ignore_unknown_fields=True)


def proto_to_json(message):
    """Converts a message to JSON, using snake_case for field names."""
    return MessageToJson(message, preserving_proto_field_name=True)


def call_endpoint(host_creds, endpoint, method, json_body, response_proto, req_context):
    # Convert json string to json dictionary, to pass to requests
    if json_body:
        json_body = json.loads(json_body)
    if method == "GET":
        response = http_request(
            host_creds=host_creds,
            endpoint=endpoint,
            method=method,
            params=json_body,
            extra_headers=req_context.get_headers(),
        )
    else:
        response = http_request(
            host_creds=host_creds,
            endpoint=endpoint,
            method=method,
            json=json_body,
            extra_headers=req_context.get_headers(),
        )
    response = verify_rest_response(response, endpoint)
    js_dict = json.loads(response.text)
    json_to_proto(js_dict=js_dict, message=response_proto)
    return response_proto



from databricks.feature_store.entities.feature_column_info import FeatureColumnInfo
from databricks.feature_store.entities.feature_table_info import FeatureTableInfo
from databricks.feature_store.entities.source_data_column_info import (
    SourceDataColumnInfo,
)
from databricks.feature_store.protos.feature_spec_pb2 import (
    FeatureSpec as ProtoFeatureSpec,
    ColumnInfo as ProtoColumnInfo,
)
from databricks.feature_store.utils import file_utils as fs_file_utils

from google.protobuf.json_format import MessageToDict, ParseDict

from mlflow.utils import file_utils as mlflow_file_utils

from typing import List, Union

from databricks.feature_store.entities.feature_spec_constants import (
    FEATURE_COLUMN_INFO,
    FEATURE_STORE,
    INPUT_COLUMNS,
    NAME,
    OUTPUT_NAME,
    SOURCE,
    SOURCE_DATA_COLUMN_INFO,
    TRAINING_DATA,
    SERIALIZATION_VERSION,
    INPUT_TABLES,
    TABLE_NAME,
    TABLE_ID,
)

# Change log for serialization version
# Please update for each serialization version.
# 1. initial.
# 2. (2021/06/16): Record feature_store_client_version to help us
# making backward compatible changes in the future.
# 3. (2021/08/25): Record table_id to handle feature table lineage stability if tables are deleted.
# 4. (2021/09/25): Record timestamp_lookup_key to handle point-in-time lookups.


class FeatureSpec:

    FEATURE_ARTIFACT_FILE = "feature_spec.yaml"
    SERIALIZATION_VERSION_NUMBER = 4

    def __init__(
        self,
        column_infos: List[Union[FeatureColumnInfo, SourceDataColumnInfo]],
        table_infos: List[FeatureTableInfo],
        workspace_id: int,
        feature_store_client_version: str,
    ):
        if not column_infos:
            raise ValueError("column_infos must be non-empty.")

        for column_info in column_infos:
            if not isinstance(column_info, (FeatureColumnInfo, SourceDataColumnInfo)):
                raise ValueError(
                    f"Expected all elements of column_infos to be instances of SourceDataColumnInfo"
                    f" or FeatureColumnInfo. '{column_info}' is of the wrong type."
                )

        self._column_infos = column_infos
        # table_infos must be present
        if table_infos is None:
            raise ValueError("table_infos must be provided.")
        # The mapping of feature table name to feature table ids.
        self._table_infos = table_infos
        self._workspace_id = workspace_id
        # The Feature Store python client version which wrote this FeatureSpec. If empty, the
        # version is <=0.3.1.
        self._feature_store_client_version = feature_store_client_version

    def __eq__(self, other):
        if not isinstance(other, FeatureSpec):
            return False
        return self.__dict__ == other.__dict__

    @property
    def column_infos(self):
        return self._column_infos

    @property
    def table_infos(self):
        return self._table_infos

    @property
    def source_data_column_infos(self) -> List[SourceDataColumnInfo]:
        return [
            col_info
            for col_info in self._column_infos
            if isinstance(col_info, SourceDataColumnInfo)
        ]

    @property
    def feature_column_infos(self) -> List[FeatureColumnInfo]:
        return [
            col_info
            for col_info in self._column_infos
            if isinstance(col_info, FeatureColumnInfo)
        ]

    @property
    def workspace_id(self):
        return self._workspace_id

    @classmethod
    def from_proto(cls, feature_spec_proto):
        # Serialization version is not deserialized from the proto as there is currently only one
        # possible version.
        column_infos = []
        for proto_column_info in feature_spec_proto.input_columns:
            if proto_column_info.HasField(SOURCE_DATA_COLUMN_INFO):
                column_infos.append(
                    SourceDataColumnInfo.from_proto(
                        proto_column_info.source_data_column_info
                    )
                )
            elif proto_column_info.HasField(FEATURE_COLUMN_INFO):
                column_infos.append(
                    FeatureColumnInfo.from_proto(proto_column_info.feature_column_info)
                )
            else:
                raise ValueError(
                    f"Expected column_info to be a SourceDataColumnInfo or FeatureColumnInfo proto "
                    f"message. '{proto_column_info}' is of the wrong proto message type."
                )
        table_infos = [
            FeatureTableInfo.from_proto(proto_table_info)
            for proto_table_info in feature_spec_proto.input_tables
        ]
        return cls(
            column_infos=column_infos,
            table_infos=table_infos,
            workspace_id=feature_spec_proto.workspace_id,
            feature_store_client_version=feature_spec_proto.feature_store_client_version,
        )

    def to_proto(self):
        proto_feature_spec = ProtoFeatureSpec()
        for column_info in self.column_infos:
            ci = ProtoColumnInfo()
            # Using CopyFrom since both the SourceDataColumnInfo and FeatureColumnInfo create their
            # own protos
            if isinstance(column_info, SourceDataColumnInfo):
                ci.source_data_column_info.CopyFrom(column_info.to_proto())
            elif isinstance(column_info, FeatureColumnInfo):
                ci.feature_column_info.CopyFrom(column_info.to_proto())
            else:
                raise ValueError(
                    f"Expected column_info to be instances of SourceDataColumnInfo"
                    f" or FeatureColumnInfo. '{column_info}' is of the wrong type."
                )
            proto_feature_spec.input_columns.append(ci)
        for table_info in self.table_infos:
            proto_feature_spec.input_tables.append(table_info.to_proto())
        proto_feature_spec.serialization_version = self.SERIALIZATION_VERSION_NUMBER
        proto_feature_spec.workspace_id = self.workspace_id
        proto_feature_spec.feature_store_client_version = (
            self._feature_store_client_version
        )
        return proto_feature_spec

    @staticmethod
    def _dict_key_by_name(column_info):
        if SOURCE_DATA_COLUMN_INFO in column_info:
            source_data = column_info[SOURCE_DATA_COLUMN_INFO]
            source_name = source_data.pop(NAME)
            source_data[SOURCE] = TRAINING_DATA
            return {source_name: source_data}
        elif FEATURE_COLUMN_INFO in column_info:
            feature_data = column_info[FEATURE_COLUMN_INFO]
            feature_data[SOURCE] = FEATURE_STORE
            return {feature_data[OUTPUT_NAME]: feature_data}
        else:
            raise ValueError(
                f"Expected column_info to be keyed by '{SOURCE_DATA_COLUMN_INFO}' and "
                f"'{FEATURE_COLUMN_INFO}'. '{column_info}' has key '{list(column_info)[0]}'."
            )

    def _to_dict(self):
        """
        Convert FeatureSpec to a writeable YAML artifact. Uses MessageToDict to convert a
        FeatureSpec proto message to a dict, then modifies the dict to be keyed by name
        :return: dict with column infos keyed by column name
        """
        # In all newer feature store clients, the unique feature tables names in input columns should always match
        # feature table names in input tables.
        unique_tables_from_feature_column_infos = set(
            [
                feature_column_info.table_name
                for feature_column_info in self.feature_column_infos
            ]
        )
        unique_tables_from_table_infos = set(
            [table_info.table_name for table_info in self.table_infos]
        )
        if unique_tables_from_feature_column_infos != unique_tables_from_table_infos:
            raise Exception(
                "Internal Error: Feature table names from input_tables "
                "does not match feature table names from input_columns."
            )

        yaml_dict = MessageToDict(self.to_proto(), preserving_proto_field_name=True)
        yaml_dict[INPUT_COLUMNS] = [
            self._dict_key_by_name(column_info)
            for column_info in yaml_dict[INPUT_COLUMNS]
        ]

        if INPUT_TABLES in yaml_dict:
            yaml_dict[INPUT_TABLES] = [
                {table_info[TABLE_NAME]: table_info}
                for table_info in yaml_dict[INPUT_TABLES]
            ]
        # For readability, place SERIALIZATION_VERSION last in the dictionary.
        reordered_keys = [k for k in yaml_dict.keys() if k != SERIALIZATION_VERSION] + [
            SERIALIZATION_VERSION
        ]
        return {k: yaml_dict[k] for k in reordered_keys}

    def save(self, path: str):
        """
        Convert spec to a YAML artifact and store at given `path` location.
        :param path: Root path to where YAML artifact is expected to be stored.
        :return: None
        """
        # TODO(ML-15922): Migrate to use mlflow file_utils once `sort_keys` argument is supported
        fs_file_utils.write_yaml(
            path, self.FEATURE_ARTIFACT_FILE, self._to_dict(), sort_keys=False
        )

    @staticmethod
    def _dict_key_by_source(column_info):
        if len(column_info) != 1:
            raise ValueError(
                f"Expected column_info dictionary to only have one key, value pair. '{column_info}'"
                f" has length {len(column_info)}."
            )
        column_name, column_data = list(column_info.items())[0]
        if not column_data:
            raise ValueError(
                f"Expected values of '{column_name}' dictionary to be non-empty."
            )
        if SOURCE not in column_data:
            raise ValueError(
                f"Expected values of column_info dictionary to include the source. No source found "
                f"for '{column_name}'."
            )
        source = column_data.pop(SOURCE)
        if source == TRAINING_DATA:
            column_data[NAME] = column_name
            return {SOURCE_DATA_COLUMN_INFO: column_data}
        elif source == FEATURE_STORE:
            return {FEATURE_COLUMN_INFO: column_data}
        else:
            raise ValueError(
                f"Expected column_info to have source of '{TRAINING_DATA}' or '{FEATURE_STORE}'. "
                f"'{column_info}' has source of '{source}'."
            )

    @classmethod
    def _from_dict(cls, spec_dict):
        """
        Convert YAML artifact to FeatureSpec. Transforms YAML artifact to dict keyed by
        source_data_column_info or feature_column_info, such that ParseDict can convert the dict to
        a proto message, and from_proto can convert the proto message to a FeatureSpec object
        :return: :py:class:`~databricks.feature_store.entities.feature_spec.FeatureSpec`
        """
        if INPUT_COLUMNS not in spec_dict:
            raise ValueError(
                f"{INPUT_COLUMNS} must be a key in {cls.FEATURE_ARTIFACT_FILE}."
            )
        if not spec_dict[INPUT_COLUMNS]:
            raise ValueError(
                f"{INPUT_COLUMNS} in {cls.FEATURE_ARTIFACT_FILE} must be non-empty."
            )
        spec_dict[INPUT_COLUMNS] = [
            cls._dict_key_by_source(column_info)
            for column_info in spec_dict[INPUT_COLUMNS]
        ]
        # For feature_spec.yaml wrote by older version of the client and for feature_spec.yaml that only has
        # source data, table_infos field may not be present
        if INPUT_TABLES not in spec_dict:
            spec_dict[INPUT_TABLES] = []
        else:
            spec_dict[INPUT_TABLES] = [
                list(input_table.values())[0] for input_table in spec_dict[INPUT_TABLES]
            ]
            # table_name in all table_infos must be unique
            unique_table_names = set(
                [input_table[TABLE_NAME] for input_table in spec_dict[INPUT_TABLES]]
            )
            if len(spec_dict[INPUT_TABLES]) != len(unique_table_names):
                raise Exception(
                    "Internal Error: Expect all table_name in input_tables to be unique. Duplicate table_name found."
                )
        return cls.from_proto(
            ParseDict(spec_dict, ProtoFeatureSpec(), ignore_unknown_fields=True)
        )

    @classmethod
    def load(cls, path: str):
        """
        Load YAML artifact and convert to FeatureSpec
        :param path: Root path to the YAML artifact.
        :return: :py:class:`~databricks.feature_store.entities.feature_spec.FeatureSpec`
        """
        spec_dict = mlflow_file_utils.read_yaml(path, cls.FEATURE_ARTIFACT_FILE)
        return cls._from_dict(spec_dict)



""" Defines the PublishDynamoDBEngine class, which defines the operations to be performed on
DynamoDB engine for publishing an offline feature table to online DynamoDB store.
"""

import abc
from typing import List, Dict, Optional

from botocore.exceptions import ClientError
from pyspark.sql import SparkSession
from pyspark.sql.functions import unix_timestamp
from pyspark.sql.types import (
    StructType,
)

from databricks.feature_store.online_store_spec import AmazonDynamoDBSpec
from databricks.feature_store.utils.dynamodb_type_utils import (
    get_timestamp_key_type,
    BASIC_DATA_TYPE_CONVERTERS,
    COMPLEX_DATA_TYPE_CONVERTERS,
    get_type_converter,
)
from databricks.feature_store.utils.dynamodb_utils import (
    get_dynamodb_resource,
    to_dynamodb_primary_key,
    PRIMARY_KEY_SCHEMA,
    PRIMARY_KEY_ATTRIBUTE_NAME_VALUE,
    ATTRIBUTE_NAME,
    KEY_TYPE,
    ATTRIBUTE_TYPE,
    DYNAMODB_STRING_TYPE,
    RANGE,
    TTL_KEY_ATTRIBUTE_NAME_VALUE,
)


class PublishDynamoDBEngine(abc.ABC):
    DYNAMODB_BILLING_MODE = "PAY_PER_REQUEST"

    def __init__(self, online_store: AmazonDynamoDBSpec, spark_session: SparkSession):
        self.spark_session = spark_session
        self.access_key_id = online_store._lookup_access_key_id_with_write_permissions()
        self.secret_access_key = (
            online_store._lookup_secret_access_key_with_write_permissions()
        )
        self.session_token = online_store._lookup_session_token_with_write_permissions()
        self.region = online_store.region
        self.table_name = online_store.table_name
        self.ttl = online_store.ttl
        self._initialize_dynamodb_client_and_resource()

    def _initialize_dynamodb_client_and_resource(self):
        """
        Initializes DynamoDB client and resource objects to perform stateless operations on DynamoDB.
        """
        self._dynamodb_resource = get_dynamodb_resource(
            access_key_id=self.access_key_id,
            secret_access_key=self.secret_access_key,
            region=self.region,
            session_token=self.session_token,
        )
        self._dynamodb_client = self._dynamodb_resource.meta.client

    def close(self):
        """
        Performs any close operations on the DynamoDB connections. DynamoDB connections are
        stateless http connections and hence does not need to be closed.
        :return:
        """
        pass

    def get_cloud_provider_unique_id(self) -> Optional[str]:
        """
        Returns cloud provider unique ID if the table exists, None otherwise
        """
        try:
            return self._dynamodb_resource.Table(self.table_name).table_arn
        except ClientError as ce:
            if ce.response["Error"]["Code"] == "ResourceNotFoundException":
                return None
            else:
                raise ce

    def create_empty_table(
        self,
        schema: StructType,
        primary_keys: List[str],
        timestamp_keys: List[str],
    ) -> str:
        """
        Creates an empty online table with the expected schema and configurations.
        (e.g. combining PKs, dropping the timestamp key, enabling ttl)
        """
        # Check: We only support a single timestamp key in DynamoDB.
        # TODO: ML-19665 add similar check in backend RPC validation.
        # TODO (ML-22021): move this validation to FeatureStoreClient.publish
        if len(timestamp_keys) > 1:
            raise ValueError(
                "Only one timestamp key is supported in DynamoDB online store."
            )

        # Check: The timestamp keys and publish mode are consistent.
        # DynamoDB tables not in window publish mode should not be created with timestamp keys,
        # as timestamp keys are logically mapped to and used in the DynamoDB composite PK RANGE column.
        if not self.is_timeseries_window_publish:
            timestamp_keys = []

        # Create the table. All validations should be prior.
        key_schema = self._get_key_schema(primary_keys, timestamp_keys)
        attribute_definitions = self._get_attribute_definitions(
            schema, primary_keys, timestamp_keys
        )
        self._dynamodb_client.create_table(
            TableName=self.table_name,
            KeySchema=key_schema,
            AttributeDefinitions=attribute_definitions,
            BillingMode=self.DYNAMODB_BILLING_MODE,
        )
        table = self._dynamodb_resource.Table(self.table_name)
        table.wait_until_exists()
        # Enable TTL attribute for this table if required.
        if self.is_timeseries_window_publish:
            self._dynamodb_client.update_time_to_live(
                TableName=self.table_name,
                TimeToLiveSpecification={
                    "Enabled": True,
                    "AttributeName": TTL_KEY_ATTRIBUTE_NAME_VALUE,
                },
            )
        # re-fetch table resource from boto3 to get the table ARN
        return self.get_cloud_provider_unique_id()

    @staticmethod
    def dynamodb_bulk_writes(
        table_name,
        df,
        primary_keys,
        access_key_id,
        secret_access_key,
        region,
        session_token,
    ):
        """
        Static method to make writes to DynamoDB using Spark foreachPatition operation.
        This method is static to avoid transferring any object state to worker. Passing object state
        fails since not all member variables are seriaizable.
        :param df:
        :param table_name:
        :param primary_keys:
        :param access_key_id:
        :param secret_access_key:
        :param region:
        :param session_token:
        :return:
        """
        schema = df.schema
        unsupported_data_types = [
            (field.name, field.dataType)
            for field in schema.fields
            if type(field.dataType) not in BASIC_DATA_TYPE_CONVERTERS
            and type(field.dataType) not in COMPLEX_DATA_TYPE_CONVERTERS
        ]

        if unsupported_data_types:
            plural = len(unsupported_data_types) > 1
            missing_cols_str = ", ".join(
                [
                    f"\n\t- {feat_name} (type: {feat_type})"
                    for (feat_name, feat_type) in unsupported_data_types
                ]
            )
            raise ValueError(
                f"Unsupported data types for DynamoDB publish. Feature{'s' if plural else ''}: {missing_cols_str}"
            )

        def row_to_items(row):
            item = {}
            primary_keys_list = []
            for index in range(len(row)):
                data_type = schema[index].dataType
                if schema[index].name in primary_keys:
                    primary_keys_list.append(
                        get_type_converter(data_type).to_dynamodb(row[index])
                    )
                else:
                    item[schema[index].name] = get_type_converter(
                        data_type
                    ).to_dynamodb(row[index])
            return {**item, **to_dynamodb_primary_key(primary_keys_list)}

        def dynamodb_bulk_write(rows):
            dynamodb_resource = get_dynamodb_resource(
                access_key_id=access_key_id,
                secret_access_key=secret_access_key,
                region=region,
                session_token=session_token,
            )
            table = dynamodb_resource.Table(table_name)

            with table.batch_writer() as batch:
                # DynamoDB batch writer handles chunking, buffering and retrying automatically.
                for row in rows:
                    batch.put_item(Item=row_to_items(row))

        df.foreachPartition(dynamodb_bulk_write)

    def write_table(
        self,
        df,
        schema: StructType,
        primary_keys: List[str],
        timestamp_keys: List[str],
    ):
        """
        Generates the expected schema and writes to an online table.
        (e.g. combining PKs, writing online-safe data types, generating ttl)
        """
        self.dynamodb_bulk_writes(
            self.table_name,
            df,
            primary_keys,
            self.access_key_id,
            self.secret_access_key,
            self.region,
            self.session_token,
        )

    def _get_key_schema(
        self, primary_keys: List[str], timestamp_keys: List[str]
    ) -> List[Dict[str, str]]:
        if not len(primary_keys):
            return []
        key_schema = [PRIMARY_KEY_SCHEMA]
        if len(timestamp_keys):
            key_schema.append({ATTRIBUTE_NAME: timestamp_keys[0], KEY_TYPE: RANGE})
        return key_schema

    def _get_attribute_definitions(
        self, schema: StructType, primary_keys: List[str], timestamp_keys: List[str]
    ) -> List[Dict[str, str]]:
        if not len(primary_keys):
            return []
        attribute_definitions = [
            {
                ATTRIBUTE_NAME: PRIMARY_KEY_ATTRIBUTE_NAME_VALUE,
                ATTRIBUTE_TYPE: DYNAMODB_STRING_TYPE,
            }
        ]
        if len(timestamp_keys):
            timestamp_key = timestamp_keys[0]
            timestamp_key_type = None
            for field in schema.fields:
                if field.name == timestamp_key:
                    timestamp_key_type = get_timestamp_key_type(field.dataType)
            attribute_definitions.append(
                {
                    ATTRIBUTE_NAME: timestamp_key,
                    ATTRIBUTE_TYPE: timestamp_key_type,
                }
            )

        return attribute_definitions

    def generate_df_with_ttl_if_required(self, df, timestamp_keys: List[str]):
        """
        Convert the timestamp column to epoch time in seconds per DynamoDB TTL expectations if TTL is defined.
        No buffer is required when generating the TTL column since the TTL is the event timestamp + the TTL duration.
        Both the event timestamp (from unix_timestamp) and DynamoDB are expected to be in UTC.
        See https://docs.aws.amazon.com/amazondynamodb/latest/developerguide/time-to-live-ttl-before-you-start.html
        """
        if self.is_timeseries_window_publish:
            ttl_seconds = int(self.ttl.total_seconds())
            ttl_column = unix_timestamp(df[timestamp_keys[0]]) + ttl_seconds
            df = df.withColumn(TTL_KEY_ATTRIBUTE_NAME_VALUE, ttl_column)
        return df

    @property
    def is_timeseries_window_publish(self):
        """
        Determine if this publish engine will publish a window of a time series dataframe.
        This is currently equivalent to if TTL is defined in the online store spec.
        """
        return self.ttl is not None



from databricks.feature_store.publish_engine.publish_sql_engine import (
    PublishSqlEngine,
)
from databricks.feature_store.publish_engine.publish_dynamodb_engine import (
    PublishDynamoDBEngine,
)
from databricks.feature_store.publish_engine.publish_mysql_engine import (
    PublishMySqlEngine,
)
from databricks.feature_store.publish_engine.publish_sql_server_engine import (
    PublishSqlServerEngine,
)
from databricks.feature_store.publish_engine.publish_cosmosdb_engine import (
    PublishCosmosDBEngine,
)

__all__ = [
    "PublishDynamoDBEngine",
    "PublishSqlEngine",
    "PublishMySqlEngine",
    "PublishSqlServerEngine",
    "PublishCosmosDBEngine",
]



import base64
from abc import abstractmethod, ABC
from datetime import date, datetime, timezone
from decimal import Decimal
from typing import Type

from pyspark.sql.types import (
    DataType,
    BooleanType,
    ShortType,
    IntegerType,
    LongType,
    FloatType,
    DoubleType,
    StringType,
    BinaryType,
    DateType,
    TimestampType,
    DecimalType,
    ArrayType,
    MapType,
)

from databricks.feature_store.utils.converter_utils import return_if_none

# Abstract converters


class CosmosDbPySparkTypeConverter(ABC):
    @staticmethod
    @abstractmethod
    def to_cosmosdb(value):
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def online_data_type() -> DataType:
        raise NotImplementedError


class CosmosDbPySparkMapKeyTypeConverter(CosmosDbPySparkTypeConverter):
    """
    Abstract class to handle conversion of Spark MapType key data types.
    Maps are stored as JSON documents in Cosmos DB, where the keys must be strings.
    """

    @staticmethod
    @abstractmethod
    def to_cosmosdb(value) -> str:
        raise NotImplementedError

    @staticmethod
    def online_data_type() -> DataType:
        return StringType()


# Converter implementations


class CosmosDbPySparkBinaryTypeConverter(CosmosDbPySparkTypeConverter):
    @staticmethod
    @return_if_none
    def to_cosmosdb(value: bytearray) -> str:
        return str(base64.b64encode(value), "utf-8")

    @staticmethod
    def online_data_type() -> DataType:
        return StringType()


class CosmosDbPySparkDateTypeConverter(CosmosDbPySparkTypeConverter):
    @staticmethod
    @return_if_none
    def to_cosmosdb(value: date) -> str:
        return value.isoformat()

    @staticmethod
    def online_data_type() -> DataType:
        return StringType()


class CosmosDbPySparkTimestampTypeConverter(CosmosDbPySparkTypeConverter):
    @staticmethod
    @return_if_none
    def to_cosmosdb(value: datetime) -> str:
        # Spark Timestamps, represented as datetime in Python, are timezone naive and assumed to be in UTC.
        # We snap the timezone to UTC and then drop the timezone info. This ensures that timezone.isoformat()
        # for timezone naive (assumed UTC) and timezone aware datetimes will generate an naive UTC ISO string.
        # The expected format is yyyy-MM-ddTHH:mm:ss.ffffff without any timezone offset information.
        dt_naive_utc = value.astimezone(timezone.utc).replace(tzinfo=None)
        return dt_naive_utc.isoformat(timespec="microseconds")

    @staticmethod
    def online_data_type() -> DataType:
        return StringType()


class CosmosDbPySparkDecimalTypeConverter(CosmosDbPySparkTypeConverter):
    @staticmethod
    @return_if_none
    def to_cosmosdb(value: Decimal) -> str:
        return str(value)

    @staticmethod
    def online_data_type() -> DataType:
        return StringType()


class CosmosDbPySparkMapKeyStringTypeConverter(CosmosDbPySparkMapKeyTypeConverter):
    @staticmethod
    def to_cosmosdb(value: str) -> str:
        return value


class CosmosDbPySparkMapKeyShortIntLongTypeConverter(
    CosmosDbPySparkMapKeyTypeConverter
):
    @staticmethod
    def to_cosmosdb(value: int) -> str:
        return str(value)


# Identity, Array, and Map converters are stateful converters


class CosmosDbPySparkIdentityConverter(CosmosDbPySparkTypeConverter):
    """
    Helper identity converter for data types that do not require conversion.
    Data types that use this should explicitly initialize a dedicated converter,
    e.g. CosmosDbPySparkBooleanTypeConverter = CosmosDbPySparkIdentityConverter(BooleanType)
    """

    def __init__(self, dt_type: Type[DataType]):
        self._dt_type = dt_type

    def to_cosmosdb(self, value):
        return value

    def online_data_type(self) -> DataType:
        return self._dt_type()


class CosmosDbPySparkArrayTypeConverter(CosmosDbPySparkTypeConverter):
    def __init__(
        self, element_converter: CosmosDbPySparkTypeConverter, contains_null: bool
    ):
        self._element_converter = element_converter
        self._contains_null = contains_null

    @return_if_none
    def to_cosmosdb(self, value: list):
        return [self._element_converter.to_cosmosdb(e) for e in value]

    def online_data_type(self) -> DataType:
        return ArrayType(
            elementType=self._element_converter.online_data_type(),
            containsNull=self._contains_null,
        )


class CosmosDbPySparkMapTypeConverter(CosmosDbPySparkTypeConverter):
    def __init__(
        self,
        key_converter: CosmosDbPySparkMapKeyTypeConverter,
        value_converter: CosmosDbPySparkTypeConverter,
        value_contains_null: bool,
    ):
        self._key_converter = key_converter
        self._value_converter = value_converter
        self._value_contains_null = value_contains_null

    @return_if_none
    def to_cosmosdb(self, value: dict) -> dict:
        return {
            self._key_converter.to_cosmosdb(k): self._value_converter.to_cosmosdb(v)
            for k, v in value.items()
        }

    def online_data_type(self) -> DataType:
        return MapType(
            keyType=self._key_converter.online_data_type(),
            valueType=self._value_converter.online_data_type(),
            valueContainsNull=self._value_contains_null,
        )


CosmosDbPySparkBooleanTypeConverter = CosmosDbPySparkIdentityConverter(BooleanType)
CosmosDbPySparkShortTypeConverter = CosmosDbPySparkIdentityConverter(ShortType)
CosmosDbPySparkIntegerTypeConverter = CosmosDbPySparkIdentityConverter(IntegerType)
CosmosDbPySparkLongTypeConverter = CosmosDbPySparkIdentityConverter(LongType)
CosmosDbPySparkFloatTypeConverter = CosmosDbPySparkIdentityConverter(FloatType)
CosmosDbPySparkDoubleTypeConverter = CosmosDbPySparkIdentityConverter(DoubleType)
CosmosDbPySparkStringTypeConverter = CosmosDbPySparkIdentityConverter(StringType)

# Helpers to get the appropriate data type converter

BASIC_DATA_TYPE_IDENTITIES = {
    BooleanType: CosmosDbPySparkBooleanTypeConverter,
    ShortType: CosmosDbPySparkShortTypeConverter,
    IntegerType: CosmosDbPySparkIntegerTypeConverter,
    LongType: CosmosDbPySparkLongTypeConverter,
    FloatType: CosmosDbPySparkFloatTypeConverter,
    DoubleType: CosmosDbPySparkDoubleTypeConverter,
    StringType: CosmosDbPySparkStringTypeConverter,
}

BASIC_DATA_TYPE_CONVERTERS = {
    **BASIC_DATA_TYPE_IDENTITIES,
    BinaryType: CosmosDbPySparkBinaryTypeConverter,
    DateType: CosmosDbPySparkDateTypeConverter,
    TimestampType: CosmosDbPySparkTimestampTypeConverter,
    DecimalType: CosmosDbPySparkDecimalTypeConverter,
}

MAP_KEY_DATA_TYPE_CONVERTERS = {
    StringType: CosmosDbPySparkMapKeyStringTypeConverter,
    ShortType: CosmosDbPySparkMapKeyShortIntLongTypeConverter,
    IntegerType: CosmosDbPySparkMapKeyShortIntLongTypeConverter,
    LongType: CosmosDbPySparkMapKeyShortIntLongTypeConverter,
}


def get_data_type_converter(data_type: DataType) -> CosmosDbPySparkTypeConverter:
    dt_type = type(data_type)
    if dt_type == ArrayType:
        return CosmosDbPySparkArrayTypeConverter(
            element_converter=get_data_type_converter(data_type.elementType),
            contains_null=data_type.containsNull,
        )
    elif dt_type == MapType:
        return CosmosDbPySparkMapTypeConverter(
            key_converter=get_map_key_data_type_converter(data_type.keyType),
            value_converter=get_data_type_converter(data_type.valueType),
            value_contains_null=data_type.valueContainsNull,
        )
    else:
        return BASIC_DATA_TYPE_CONVERTERS[dt_type]


def get_map_key_data_type_converter(
    data_type: DataType,
) -> CosmosDbPySparkMapKeyTypeConverter:
    dt_type = type(data_type)
    if dt_type not in MAP_KEY_DATA_TYPE_CONVERTERS:
        raise ValueError(
            f"Unsupported map key data type for Cosmos DB publish: {dt_type}"
        )
    return MAP_KEY_DATA_TYPE_CONVERTERS[dt_type]



import uuid

import pyspark.sql.functions as f
from pyspark.sql import Window

# Publish latest feature values for tables with timestamp key.
# aka. feature values with the latest timestamp for each unique primary keys combination.
from databricks.feature_store.entities.online_store_detailed import OnlineStoreDetailed
from databricks.feature_store.online_store_spec import OnlineStoreSpec
from databricks.feature_store.online_store_publish_client import is_ttl_spec


def get_latest_snapshot(df, primary_keys, timestamp_keys):
    timestamp_key = timestamp_keys[0]
    # temporary timestamp rank key used for filtering, add random suffix to avoid column collision
    timestamp_rank_key = f"{timestamp_key}_rank_{uuid.uuid4().hex}"
    # partition DataFrame by primary keys and rank by timestamp key in descending order
    timestamp_rank_window = Window.partitionBy(primary_keys).orderBy(
        f.col(timestamp_key).desc()
    )
    # filter out the highest rank (latest snapshot) and drop the timestamp rank key
    return (
        df.withColumn(timestamp_rank_key, f.rank().over(timestamp_rank_window))
        .filter(f.col(timestamp_rank_key) == 1)
        .drop(timestamp_rank_key)
    )


def update_online_store_spec_sticky_ttl(
    online_store_spec: OnlineStoreSpec, online_store_detailed: OnlineStoreDetailed
):
    """
    Resolves the TTL and returns an updated OnlineStoreSpec. The OnlineStoreSpec TTL is used if defined.
    Otherwise, a sticky TTL is inherited from the OnlineStoreDetailed.
    """
    # Nothing needs to be done if the OnlineStoreSpec doesn't support TTL.
    if not is_ttl_spec(online_store_spec):
        return online_store_spec

    resolved_ttl = (
        online_store_spec.ttl
        if online_store_spec.ttl
        else online_store_detailed.additional_metadata.ttl
    )
    return online_store_spec.clone(**{"ttl": resolved_ttl})



import abc
import copy
import logging
from typing import Union, List, Dict

from mlflow.utils.annotations import deprecated

from databricks.feature_store.online_store_spec.online_store_properties import (
    TYPE,
    HOSTNAME,
    PORT,
    DATABASE_NAME,
    TABLE_NAME,
    USER,
    PASSWORD,
    DRIVER_NAME,
    READ_SECRET_PREFIX,
    WRITE_SECRET_PREFIX,
    EXPLICIT_CREDENTIAL_PARAMS,
)
from databricks.feature_store.utils.utils import is_empty, _get_dbutils as get_dbutils

_logger = logging.getLogger(__name__)


class OnlineStoreSpec(abc.ABC):
    """
    Parent class for all types of :class:`OnlineStoreSpec` objects.

    Abstract base class for classes that specify the online store to publish to.

    If `database_name` and `table_name` are not provided,
    :meth:`FeatureStoreClient.publish_table <databricks.feature_store.client.FeatureStoreClient.publish_table>`
    will use the offline store's database and table
    names.

    To use a different database and table name in the online store, provide values for
    both `database_name` and `table_name` arguments.

    The JDBC driver can be customized with the optional ``driver_name`` argument.
    Otherwise, a default is used.

    Strings in the primary key should not exceed 100 characters.

    The online database should already exist.

    .. note::

       It is strongly suggested (but not required), to provide read-only database credentials via
       the ``read_secret_prefix`` in order to grant the least amount of database access
       privileges to the served model.  When providing a ``read_secret_prefix``, the secrets must
       exist in the scope name using the expected format,
       otherwise :meth:`publish_table <databricks.feature_store.client.FeatureStoreclient.publish_table>` will return an error.

    :param hostname: Hostname to access online store. The database hostname cannot be changed. Subsequent publish
      calls to the same online store must provide the same hostname.
    :param port: Port number to access online store. The database port cannot be changed. Subsequent publish
      calls to the same online store must provide the same port.
    :param user: Username that has write access to the online store. **Deprecated** as of version 0.6.0.
      Use ``write_secret_prefix`` instead.
    :param password: Password to access the online store. **Deprecated** as of version 0.6.0.
      Use ``write_secret_prefix`` instead.
    :param database_name: Database name.
    :param table_name: Table name.
    :param driver_name: Name of custom JDBC driver to access the online store.
    :param read_secret_prefix: The secret scope name and secret key name prefix where read-only online store
      credentials are stored. These credentials will be used during online feature serving to connect to the
      online store from the served model. The format of this parameter should be ``${scope-name}/${prefix}``,
      which is the name of the secret scope, followed by a ``/``, followed by the secret key name prefix. The
      scope passed in must contain the following keys and corresponding values:

      * ``${prefix}-user`` where ``${prefix}`` is the value passed into this function. For example if this
        function is called with ``datascience/staging``, the ``datascience`` secret scope should contain the
        secret named ``staging-user``, which points to a secret value with the database username for the
        online store.

      * ``${prefix}-password`` where ``${prefix}`` is the value passed into this function. For example if this
        function is called with ``datascience/staging``, the ``datascience`` secret scope should contain the
        secret named ``staging-password``, which points to a secret value with the database password for the
        online store.

      Once the read_secret_prefix is set for an online store, it cannot be changed.

    :param write_secret_prefix: The secret scope name and secret key name prefix where read-write online store
      credentials are stored. These credentials will be used to connect to the online store to publish
      features. If ``user`` and ``password`` are passed, this field must be ``None``, or an exception will be raised.
      The format of this parameter should be ``${scope-name}/${prefix}``, which is the name of the secret scope,
      followed by a ``/``, followed by the secret key name prefix. The scope passed in must contain the following
      keys and corresponding values:

      * ``${prefix}-user`` where ``${prefix}`` is the value passed into this function.  For example if this
        function is called with ``datascience/staging``, the ``datascience`` secret scope should contain the
        secret named ``staging-user``, which points to a secret value with the database username for the
        online store.

      * ``${prefix}-password`` where ``${prefix}`` is the value passed into this function.  For example if this
        function is called with ``datascience/staging``, the ``datascience`` secret scope should contain the
        secret named ``staging-password``, which points to a secret value with the database password for the
        online store.

    """

    # TODO (ML-23105): Remove explicit parameters for MLR 12.0.
    @abc.abstractmethod
    def __init__(
        self,
        _type,
        hostname: [str, None] = None,
        port: [int, None] = None,
        user: Union[str, None] = None,
        password: Union[str, None] = None,
        database_name: Union[str, None] = None,
        table_name: Union[str, None] = None,
        driver_name: Union[str, None] = None,
        read_secret_prefix: Union[str, None] = None,
        write_secret_prefix: Union[str, None] = None,
        _internal_properties: Union[Dict[str, str], None] = None,
    ):
        self._properties = {
            TYPE: _type,
            HOSTNAME: hostname,
            PORT: port,
            DATABASE_NAME: database_name,
            TABLE_NAME: table_name,
            USER: user,
            PASSWORD: password,
            DRIVER_NAME: driver_name,
            READ_SECRET_PREFIX: read_secret_prefix,
            WRITE_SECRET_PREFIX: write_secret_prefix,
        }
        if _internal_properties:
            self._properties.update(_internal_properties)
        self._validate_credentials()
        self._warn_on_explicit_credentials()

    def _validate_credentials(self):
        """
        Validate that the expected credentials were provided and are unambiguous.
        Subclasses with different behavior should override this.
        """
        # Validate that the user passed either a user/pass combination or a write_secret_prefix, else throw an error
        # because otherwise there are no database credentials to use
        # TODO (ML-19653): Align the allowed credentials and error messages
        if self.write_secret_prefix is None and (
            self.user is None or self.password is None
        ):
            raise Exception(
                "Use either 'user'/'password' combination or 'write_secret_prefix'."
            )

        # Validate that user didn't pass user/pass AND write_secret_prefix, or else throw an error because this
        # is ambiguous
        if not is_empty(self.user) and self.write_secret_prefix is not None:
            raise Exception("Use either 'user' or 'write_secret_prefix', but not both.")
        if not is_empty(self.password) and self.write_secret_prefix is not None:
            raise Exception(
                "Use either 'password' or 'write_secret_prefix', but not both."
            )

    def _warn_on_explicit_credentials(self):
        """
        Warn if any explicit credential parameters are used as they are deprecated.
        """
        explicit_credentials = [
            f'"{p}"'
            for p in EXPLICIT_CREDENTIAL_PARAMS
            if self._properties.get(p, None)
        ]
        if explicit_credentials:
            # Parameters use underscore "_". while secrets and the internal constants use dash "-"
            params_str = ", ".join(explicit_credentials).replace("-", "_")
            _logger.warning(
                f'The explicit credential parameters {params_str} is deprecated. Use "write_secret_prefix" instead.'
            )

    @property
    def type(self):
        """Type of the online store."""
        return self._properties[TYPE]

    @property
    def table_name(self):
        """Table name."""
        return self._properties[TABLE_NAME]

    @property
    @deprecated(alternative="write_secret_prefix", since="v0.6.0")
    def user(self):
        """Username that has access to the online store.

        Property will be empty if ``write_secret_prefix`` argument was used.
        """
        return self._properties[USER]

    @property
    @deprecated(alternative="write_secret_prefix", since="v0.6.0")
    def password(self):
        """Password to access the online store.

        Property will be empty if ``write_secret_prefix`` argument was used.
        """
        return self._properties[PASSWORD]

    @property
    def driver(self):
        """Name of the custom JDBC driver to access the online store."""
        return self._properties[DRIVER_NAME]

    @property
    def read_secret_prefix(self):
        """
        Prefix for read access to online store.

        Name of the secret scope and prefix that contains the username and password to access
        the online store with read-only credentials.

        See the ``read_secret_prefix`` parameter description for details.
        """
        return self._properties[READ_SECRET_PREFIX]

    @property
    def write_secret_prefix(self):
        """
        Secret prefix that contains online store login info.

        Name of the secret scope and prefix that contains the username and password to access
        the online store with read/write credentials.
        See the ``write_secret_prefix`` parameter description for details.
        """
        return self._properties[WRITE_SECRET_PREFIX]

    @property
    def cloud(self):
        """Cloud provider where this online store is located."""
        raise NotImplementedError

    @property
    def store_type(self):
        """Store type."""
        raise NotImplementedError

    def auth_type(self):
        """Publish Auth type."""
        raise NotImplementedError

    def clone(self, **kwargs):
        """Clone a feature spec."""
        new_spec = copy.deepcopy(self)
        new_spec._properties.update(kwargs)
        return new_spec

    def non_secret_properties(self):
        """
        TODO (ML-21692): Determine if this function is required and update for use if necessary.

        Dictionary of non-secret properties.
        """
        keys = [
            TYPE,
            HOSTNAME,
            PORT,
            DATABASE_NAME,
            TABLE_NAME,
        ]
        return {key: self._properties[key] for key in keys}

    def _lookup_user_with_write_permissions(self) -> str:
        """
        Username that has write access to the online store, based on either being provided directly via the constructor
        or by resolving via write_secret_prefix and dbutils.

        This contains only the username as stored in the secret.  If you need the authorization user for the
        purposes of putting into a jdbc connection string, which may contain more than just the username,
        call _lookup_jdbc_auth_user_with_write_permissions() instead.

        WARNING: do not hold onto the returned secret for longer than necessary, for example saving in
        data structures, files, other persistent backends.  Use it only for directly accessing resources
        and then allow the Python VM to remove the reference as soon as it's out of scope.
        """
        return self._lookup_secret_with_write_permissions(USER)

    def _lookup_password_with_write_permissions(self) -> str:
        """
        Password that has write access to the online store, based on either being provided directly via the constructor
        or by resolving via write_secret_prefix and dbutils.

        WARNING: do not hold onto the returned secret for longer than necessary, for example saving in
        data structures, files, other persistent backends.  Use it only for directly accessing resources
        and then allow the Python VM to remove the reference as soon as it's out of scope.
        """
        return self._lookup_secret_with_write_permissions(PASSWORD)

    def _valid_secret_suffixes(self) -> List[str]:
        """
        List of valid secret suffixes.
        Defaults to "user" and "password," but subclasses with different behavior should override this.
        """
        return [USER, PASSWORD]

    def _expected_secret_suffixes(self) -> List[str]:
        """
        List of expected secret suffixes.
        Defaults to "user" and "password," but subclasses with different behavior should override this.
        """
        return [USER, PASSWORD]

    def _lookup_secret_with_write_permissions(self, secret) -> str:
        """
        Lookup secret (either user or password) that has write permissions to the online store.

        WARNING: do not hold onto the returned secrets for longer than necessary, for example saving in
        data structures, files, other persistent backends.  Use it only for directly accessing resources
        and then allow the VM to clean up the variable in memory.
        """

        # Ensure a valid secret was passed
        if secret not in self._valid_secret_suffixes():
            raise Exception("Internal error: invalid secret passed")

        # Validate the read_secret_prefix to catch missing secrets as early as possible.
        # TODO (ML-21737): Determine if this validation should be done earlier.
        self._validate_read_secrets()

        # If the user passed in the secret directly via the constructor rather than via WRITE_SECRET_PREFIX, return it
        if not is_empty(self._properties.get(secret)):
            return self._properties[secret]

        # Otherwise the user must have passed the secret via WRITE_SECRET_PREFIX, so resolve the secret and return it
        return self._lookup_write_secret_for_suffix(secret)

    def _lookup_jdbc_auth_user_with_write_permissions(self) -> str:
        """
        Generate the user authentication string used in the JDBC URL.  The default behavior is "user", but subclasses
        should override this if the particular online store requires customization.  For example, on AzureMysql
        it requires the format: "user@server"
        """
        return self._lookup_user_with_write_permissions()

    def _lookup_write_secret_for_suffix(self, suffix) -> Union[str, None]:
        """
        Resolve and validate the secrets referenced in the write_secret_prefix via dbutils for the given suffix.
        Valid suffix values depend on the online store, and are defined by `_valid_secret_prefixes`.

        This must happen during the construction of the jdbc url rather than during construction of the OnlineStoreSpec object
        because the user may modify the underlying secrets being pointed to by WRITE_SECRET_PREFIX, and we need to
        validate the latest secret values rather than a snapshot of the values at the time OnlineStoreSpec was created.

        WARNING: do not hold onto the returned secrets for longer than necessary, for example saving in
        data structures, files, other persistent backends.  Use it only for directly accessing resources
        and then allow the VM to clean up the variable in memory.

        Returns the secret corresponding to either the user or password, depending on the given suffix.
        """

        # If write_secret_prefix is None, there is nothing to do
        if self._properties[WRITE_SECRET_PREFIX] is None:
            return None

        # Ensure a valid suffix was passed
        if suffix not in self._valid_secret_suffixes():
            raise Exception("Internal error: invalid suffix passed")

        # Get a reference to dbutils, which only works within a notebook or notebook job context
        dbutils = get_dbutils()

        # Use dbutils to lookup write_scope_secrets and return to the caller.  Propagate the exception if either
        # the scope or either secret does not exist in the secret store.
        write_creds_scope, write_creds_prefix = _parse_scope_prefix(
            self._properties[WRITE_SECRET_PREFIX]
        )
        return dbutils.secrets.get(
            scope=write_creds_scope,
            key=f"{write_creds_prefix}-{suffix}",
        )

    def _validate_read_secrets(self):
        """
        Validate that the READ_SECRET_PREFIX contains a valid scope with the expected secrets.
        """

        # If read_secret_prefix is None, there is nothing to do
        if self._properties[READ_SECRET_PREFIX] is None:
            return

        # Get a reference to dbutils, which only works within a notebook or notebook job context
        dbutils = get_dbutils()

        # Use dbutils to lookup read_scope_secrets and store in properties, or throw an error if the scope or secrets
        # do not exist.  The actual secrets are discarded since the lookup is performed only for validation purposes.
        read_creds_scope, read_creds_prefix = _parse_scope_prefix(
            self._properties[READ_SECRET_PREFIX]
        )
        # Check all the expected suffixes are present in the read secrets
        for secret_suffix in self._expected_secret_suffixes():
            dbutils.secrets.get(
                scope=read_creds_scope,
                key=f"{read_creds_prefix}-{secret_suffix}",
            )


def _parse_scope_prefix(scope_prefix):
    """
    Parse the secret scope + prefix into its components and do validation.

    foo/bar -> ["foo", "bar"]
    """

    format_message = (
        "The format of this parameter should be ${scope}/${prefix}, which is the name of the "
        "secret scope, followed by a /, followed by a user-defined prefix that is used to "
        "determine the secret key to be used."
    )

    if scope_prefix is None or len(scope_prefix) == 0:
        raise Exception(format_message)

    if "/" not in scope_prefix:
        raise Exception(format_message)

    split_vals = scope_prefix.split("/")

    if len(split_vals) != 2:
        raise Exception(format_message)

    return split_vals



from datetime import timedelta
from typing import Union, List, Optional

from mlflow.utils.annotations import deprecated
from pyspark.sql.utils import IllegalArgumentException

from databricks.feature_store.entities.cloud import Cloud
from databricks.feature_store.entities.store_type import StoreType
from databricks.feature_store.online_store_spec import (
    OnlineStoreSpec,
)
from databricks.feature_store.online_store_spec.online_store_properties import (
    AWS_DYNAMODB,
    REGION,
    SESSION_TOKEN,
    SECRET_ACCESS_KEY,
    ACCESS_KEY_ID,
    TTL,
    SECRETS,
    ROLE,
)
from databricks.feature_store.utils.utils import is_empty
import logging

_logger = logging.getLogger(__name__)


class AmazonDynamoDBSpec(OnlineStoreSpec):
    """
    This :class:`OnlineStoreSpec <databricks.feature_store.online_store_spec.online_store_spec.OnlineStoreSpec>`
    implementation is intended for publishing features to Amazon DynamoDB.

    If `table_name` is not provided,
    :meth:`FeatureStoreClient.publish_table <databricks.feature_store.client.FeatureStoreClient.publish_table>`
    will use the offline store's database and table name combined as the online table name.

    To use a different table name in the online store, provide a value for the `table_name` argument.

    The expected read or write secrets for DynamoDB for a given ``{prefix}`` string are
    ``${prefix}-access-key-id``, ``${prefix}-secret-access-key``, and ``${prefix}-session-token``.

    If none of the access_key_id, secret_access_key, and write_secret_prefix are passed,
    the instance profile attached to the cluster will be used to write to DynamoDB.

    .. note::

      AmazonDynamoDBSpec is available in version >= 0.3.8.

      Instance profile based writes are available in version >= 0.4.1.

    :param region: Region to access online store.
    :param access_key_id: Access key ID that has access to the online store. **Deprecated** as of version 0.6.0.
      Use ``write_secret_prefix`` instead.
    :param secret_access_key: Secret access key to access the online store. **Deprecated** as of version 0.6.0.
      Use ``write_secret_prefix`` instead.
    :param session_token: Session token to access the online store. **Deprecated** as of version 0.6.0.
      Use ``write_secret_prefix`` instead.
    :param table_name: Table name.
    :param read_secret_prefix: Prefix for read secret.
    :param write_secret_prefix: Prefix for write secret.
    :param ttl: The time to live for data published to the online store. This attribute is only applicable when
      publishing time series feature tables. If the time to live is specified for a time series table,
      :meth:`FeatureStoreClient.publish_table` will publish a window of data instead of the latest snapshot.
    """

    # TODO (ML-23105): Remove explicit parameters for MLR 12.0.
    def __init__(
        self,
        *,
        region: Union[str, None],
        access_key_id: Union[str, None] = None,
        secret_access_key: Union[str, None] = None,
        session_token: Union[str, None] = None,
        table_name: Union[str, None] = None,
        read_secret_prefix: Union[str, None] = None,
        write_secret_prefix: Union[str, None] = None,
        ttl: Optional[timedelta] = None,
    ):
        """Initialize AmazonDynamoDBSpec object."""
        super().__init__(
            AWS_DYNAMODB,
            table_name=table_name,
            read_secret_prefix=read_secret_prefix,
            write_secret_prefix=write_secret_prefix,
            _internal_properties={
                REGION: region,
                ACCESS_KEY_ID: access_key_id,
                SECRET_ACCESS_KEY: secret_access_key,
                SESSION_TOKEN: session_token,
                TTL: ttl,
            },
        )

    @property
    @deprecated(alternative="write_secret_prefix", since="v0.6.0")
    def access_key_id(self):
        """
        Access key ID that has access to the online store.
        Property will be empty if ``write_secret_prefix`` or the
        instance profile attached to the cluster are intended to be used.
        """
        return self._properties[ACCESS_KEY_ID]

    @property
    @deprecated(alternative="write_secret_prefix", since="v0.6.0")
    def secret_access_key(self):
        """
        Secret access key to access the online store.
        Property will be empty if ``write_secret_prefix`` or the
        instance profile attached to the cluster are intended to be used.
        """
        return self._properties[SECRET_ACCESS_KEY]

    @property
    @deprecated(alternative="write_secret_prefix", since="v0.6.0")
    def session_token(self):
        """
        Session token to access the online store.
        Property will be empty if ``write_secret_prefix`` or the
        instance profile attached to the cluster are intended to be used.
        """
        return self._properties[SESSION_TOKEN]

    @property
    def cloud(self):
        """Define the cloud property for the data store."""
        return Cloud.AWS

    @property
    def store_type(self):
        """Define the data store type."""
        return StoreType.DYNAMODB

    @property
    def region(self):
        """Region to access the online store."""
        return self._properties[REGION]

    @property
    def ttl(self) -> Optional[timedelta]:
        """Time to live attribute for the online store."""
        return self._properties[TTL]

    def auth_type(self):
        """Publish Auth type."""
        return ROLE if self._is_role_based() else SECRETS

    def _is_role_based(self) -> bool:
        return (
            is_empty(self.access_key_id)
            and is_empty(self.secret_access_key)
            and is_empty(self.session_token)
            and is_empty(self.write_secret_prefix)
        )

    def _lookup_access_key_id_with_write_permissions(self) -> str:
        """
        Access key ID that has write access to the online store, resolved through the write_secret_prefix and dbutils.

        WARNING: do not hold onto the returned secret for longer than necessary, for example saving in
        data structures, files, other persistent backends. Use it only for directly accessing resources
        and then allow the Python VM to remove the reference as soon as it's out of scope.
        """
        return self._lookup_secret_with_write_permissions(ACCESS_KEY_ID)

    def _lookup_secret_access_key_with_write_permissions(self) -> str:
        """
        Secret access key that has write access to the online store, resolved through the write_secret_prefix and dbutils.

        WARNING: do not hold onto the returned secret for longer than necessary, for example saving in
        data structures, files, other persistent backends. Use it only for directly accessing resources
        and then allow the Python VM to remove the reference as soon as it's out of scope.
        """
        return self._lookup_secret_with_write_permissions(SECRET_ACCESS_KEY)

    def _lookup_session_token_with_write_permissions(self) -> Union[str, None]:
        """
        Optional session token that has write access to the online store, resolved through the write_secret_prefix and dbutils.

        WARNING: do not hold onto the returned secret for longer than necessary, for example saving in
        data structures, files, other persistent backends. Use it only for directly accessing resources
        and then allow the Python VM to remove the reference as soon as it's out of scope.
        """
        # An invalid scope/key throws: "IllegalArgumentException: Secret does not exist with scope: X and key: Y"
        try:
            return self._lookup_secret_with_write_permissions(SESSION_TOKEN)
        except IllegalArgumentException:
            return None

    def _validate_credentials(self):
        """
        Validate that the expected credentials were provided and are unambiguous.
        """
        # Validate that the user passed either both (access_key_id, secret_access_key) or a write_secret_prefix,
        # else assume the user intends to use attached instance profile in the cluster.
        # TODO (ML-19653): Align the allowed credentials and error messages

        # If all write secret fields are empty, assume using role based publish.
        if self._is_role_based():
            _logger.info(
                "No explicit credentials for write provided. "
                "Instance profile attached to the cluster will be used."
            )
            return
        # Validate that user didn't pass in any of (access_key_id, secret_access_key, session_token)
        # AND write_secret_prefix, else throw an error because this is ambiguous
        if self.write_secret_prefix is not None:
            if not is_empty(self.access_key_id):
                raise Exception(
                    "Use either 'access_key_id' or 'write_secret_prefix', but not both."
                )
            if not is_empty(self.secret_access_key):
                raise Exception(
                    "Use either 'secret_access_key' or 'write_secret_prefix', but not both."
                )
            if not is_empty(self.session_token):
                raise Exception(
                    "Use either 'session_token' or 'write_secret_prefix', but not both."
                )
        else:
            if is_empty(self.access_key_id) != is_empty(self.secret_access_key):
                raise Exception(
                    "Both 'access_key_id' and 'secret_access_key' needs to be provided."
                )
            if (
                is_empty(self.access_key_id)
                and is_empty(self.secret_access_key)
                and not is_empty(self.session_token)
            ):
                raise Exception(
                    "'session_token' can only be used along with both 'access_key_id' and 'secret_access_key'."
                )

    def _valid_secret_suffixes(self) -> List[str]:
        """
        List of valid secret suffixes.
        """
        return [ACCESS_KEY_ID, SECRET_ACCESS_KEY, SESSION_TOKEN]

    def _expected_secret_suffixes(self) -> List[str]:
        """
        List of expected secret suffixes.
        """
        return [ACCESS_KEY_ID, SECRET_ACCESS_KEY]



import abc


class ColumnInfo:
    def __eq__(cls, other):
        if not isinstance(other, cls.__class__):
            return False
        return cls.__dict__ == other.__dict__

    @property
    @abc.abstractmethod
    def output_name(cls):
        pass



# Constants defining feature table property names.
PRIMARY_KEYS = "databricks.feature_store.primary_keys"
PARTITION_COLUMNS = "databricks.feature_store.partition_columns"
DESCRIPTION = "databricks.feature_store.description"
ONLINE_STORES = "databricks.feature_store.online_stores"

ALL_PROPERTIES = [PRIMARY_KEYS, PARTITION_COLUMNS, DESCRIPTION, ONLINE_STORES]



import logging

from typing import List, Dict, Optional

from databricks.feature_store.entities.consumer import Consumer
from databricks.feature_store.entities.feature_table import FeatureTable
from databricks.feature_store.entities.feature import Feature
from databricks.feature_store.entities.tag import Tag
from databricks.feature_store.entities.online_store_detailed import OnlineStoreDetailed
from databricks.feature_store.entities.online_store_metadata import OnlineStoreMetadata
from databricks.feature_store.entities._permission_level import _PermissionLevel
from databricks.feature_store.entities.online_feature_table import OnlineFeatureTable
from databricks.feature_store.utils.request_context import RequestContext
from databricks.feature_store.entities.store_type import StoreType

from databricks.feature_store.api.proto.feature_catalog_pb2 import (
    FeatureStoreService,
    CreateFeatureTable,
    GetFeatureTable,
    PublishFeatureTable,
    CreateFeatures,
    GetFeatures,
    UpdateFeature,
    KeySpec,
    DeleteFeatureTable,
    AddDataSources,
    DeleteDataSources,
    AddProducer,
    AddConsumer,
    GetConsumers,
    ConsumedFeatures,
    Notebook,
    Job,
    FeatureTableFeatures,
    GetModelServingMetadata,
    ProducerAction,
    Tag as ProtoTag,
    SetTags,
    DeleteTags,
    GetTags,
    GetOnlineStore,
)

from databricks.feature_store.utils.rest_utils import (
    call_endpoint,
    extract_api_info_for_service,
    _REST_API_PATH_PREFIX,
    proto_to_json,
    get_error_code,
)
from mlflow.protos.databricks_pb2 import ErrorCode, RESOURCE_DOES_NOT_EXIST
from mlflow.utils import databricks_utils

_METHOD_TO_INFO = extract_api_info_for_service(
    FeatureStoreService, _REST_API_PATH_PREFIX
)

_logger = logging.getLogger(__name__)


class CatalogClient:
    """
    This provides the client interface to the backend feature catalog service running in the Databricks Control Plane.

    The catalog client should be reserved for low-level catalog operations and not contain any business logic
    that is unrelated to the catalog itself (for example, calling other Databricks backend services).  If you need
    additional business logic, consider using the CatalogClientHelper instead.
    """

    def __init__(self, get_host_creds, feature_store_uri: Optional[str] = None):
        """
        Catalog client for the Feature Store client. Takes in an optional parameter to identify the remote workspace
        for multi-workspace Feature Store.
        :param feature_store_uri: An URI of the form ``databricks://<scope>.<prefix>`` that identifies the credentials
          of the intended Feature Store workspace. Throws an error if specified but credentials were not found.
        """
        self._get_host_creds = lambda: get_host_creds(feature_store_uri)
        self._local_host, self._local_workspace_id = self._get_local_workspace_info()
        (
            self._feature_store_workspace_host,
            self._feature_store_workspace_id,
        ) = self._get_feature_store_workspace_info(feature_store_uri)

    @property
    def feature_store_workspace_id(self) -> int:
        return self._feature_store_workspace_id

    def _get_local_workspace_info(self) -> (str, int):
        local_host, workspace_id = databricks_utils.get_workspace_info_from_dbutils()
        return local_host, self._parse_workspace_id(workspace_id)

    def _get_feature_store_workspace_info(
        self, feature_store_uri: Optional[str] = None
    ) -> (str, int):
        if feature_store_uri:
            # Retrieve the remote hostname and workspace ID stored in the secret scope by the user.
            (
                remote_hostname,
                remote_workspace_id,
            ) = databricks_utils.get_workspace_info_from_databricks_secrets(
                feature_store_uri
            )
            if not remote_workspace_id:
                raise ValueError(
                    f"'FeatureStoreClient' was initialized with 'feature_store_uri' argument "
                    f"for multi-workspace usage, but the remote Feature Store workspace ID was not "
                    f"found at URI {feature_store_uri}."
                )

            if not remote_hostname:
                raise ValueError(
                    f"'FeatureStoreClient' was initialized with 'feature_store_uri' argument "
                    f"for multi-workspace usage, but the remote Feature Store hostname was not "
                    f"found at URI {feature_store_uri}."
                )

            return remote_hostname, self._parse_workspace_id(remote_workspace_id)
        else:
            return self._local_host, self._local_workspace_id

    @staticmethod
    def _parse_workspace_id(workspace_id) -> int:
        if workspace_id:
            try:
                workspace_id = int(workspace_id)
            except (ValueError, TypeError):
                raise ValueError("Internal Error: Workspace ID was not found.")
        return workspace_id

    def _call_endpoint(self, api, json_body, req_context: RequestContext):
        endpoint, method = _METHOD_TO_INFO[api]
        response_proto = api.Response()
        return call_endpoint(
            self._get_host_creds(),
            endpoint,
            method,
            proto_to_json(json_body),
            response_proto,
            req_context,
        )

    def _get_feature_table(self, feature_table: str, req_context: RequestContext):
        req_body = GetFeatureTable(name=feature_table)
        return self._call_endpoint(GetFeatureTable, req_body, req_context)

    # CRUD API to call Feature Catalog
    def create_feature_table(
        self,
        feature_table: str,
        partition_key_spec,
        primary_key_spec,
        timestamp_key_spec,
        description: str,
        is_imported: str,
        req_context: RequestContext,
    ):
        req_body = CreateFeatureTable(
            name=feature_table,
            primary_keys=([key_spec.to_proto() for key_spec in primary_key_spec]),
            partition_keys=([key_spec.to_proto() for key_spec in partition_key_spec]),
            timestamp_keys=([key_spec.to_proto() for key_spec in timestamp_key_spec]),
            description=description,
            is_imported=is_imported,
        )
        response_proto = self._call_endpoint(CreateFeatureTable, req_body, req_context)
        return FeatureTable.from_proto(response_proto.feature_table)

    def get_feature_table(self, feature_table: str, req_context: RequestContext):
        response_proto = self._get_feature_table(feature_table, req_context)
        return FeatureTable.from_proto(response_proto.feature_table)

    def feature_table_exists(self, name: str, req_context: RequestContext):
        """
        Checks whether the feature table exists.

        This CatalogClient method is built on top of the feature-tables/get endpoint. There is no
        dedicated endpoint for feature_table_exists.
        """
        try:
            self.get_feature_table(name, req_context)
        except Exception as e:
            if get_error_code(e) == ErrorCode.Name(RESOURCE_DOES_NOT_EXIST):
                return False
            raise e
        return True

    def can_write_to_catalog(
        self, feature_table_name: str, req_context: RequestContext
    ):
        """
        Checks whether the user has write permission to feature catalog.

        This CatalogClient method is built on top of the feature-tables/get endpoint. There is no
        dedicated endpoint for can_write_to_catalog.
        """
        response_proto = self._get_feature_table(feature_table_name, req_context)
        return _PermissionLevel.can_write_to_catalog(
            response_proto.feature_table.permission_level
        )

    def publish_feature_table(
        self,
        feature_table: str,
        online_store_metadata: OnlineStoreMetadata,
        features: List[str],
        req_context: RequestContext,
    ):
        req_body = PublishFeatureTable(
            feature_table=feature_table,
            online_table=online_store_metadata.online_table,
            cloud=online_store_metadata.cloud,
            store_type=online_store_metadata.store_type,
            read_secret_prefix=online_store_metadata.read_secret_prefix,
            write_secret_prefix=online_store_metadata.write_secret_prefix,
            features=features,
        )
        if online_store_metadata.store_type == StoreType.MYSQL:
            req_body.mysql_metadata.CopyFrom(
                online_store_metadata.additional_metadata.to_proto()
            )
        elif online_store_metadata.store_type == StoreType.SQL_SERVER:
            req_body.sql_server_metadata.CopyFrom(
                online_store_metadata.additional_metadata.to_proto()
            )
        elif online_store_metadata.store_type == StoreType.DYNAMODB:
            req_body.dynamodb_metadata.CopyFrom(
                online_store_metadata.additional_metadata.to_proto()
            )
        elif online_store_metadata.store_type == StoreType.COSMOSDB:
            req_body.cosmosdb_metadata.CopyFrom(
                online_store_metadata.additional_metadata.to_proto()
            )
        else:
            raise TypeError(
                f"Unsupported online store metadata type {online_store_metadata.additional_metadata}"
            )

        response_proto = self._call_endpoint(PublishFeatureTable, req_body, req_context)
        return OnlineStoreDetailed.from_proto(response_proto.online_store)

    def create_features(
        self, feature_table: str, feature_specs, req_context: RequestContext
    ):
        req_body = CreateFeatures(
            feature_table=feature_table,
            features=[key_spec.to_proto() for key_spec in feature_specs],
        )
        self._call_endpoint(CreateFeatures, req_body, req_context)

    def get_features(self, feature_table: str, req_context: RequestContext):
        all_features = []
        page_token = None
        while True:
            # Use default max_results
            req_body = GetFeatures(feature_table=feature_table, page_token=page_token)
            response_proto = self._call_endpoint(GetFeatures, req_body, req_context)
            all_features += [
                Feature.from_proto(feature) for feature in response_proto.features
            ]
            page_token = response_proto.next_page_token
            if not page_token:
                break
        return all_features

    def update_feature(
        self,
        feature_table: str,
        feature: str,
        data_type: str,
        req_context: RequestContext,
    ):
        req_body = UpdateFeature(
            feature_table=feature_table, name=feature, data_type=data_type.upper()
        )
        response_body = self._call_endpoint(UpdateFeature, req_body, req_context)
        return Feature.from_proto(response_body.feature)

    def delete_feature_table(
        self, feature_table: str, req_context: RequestContext, dry_run=False
    ):
        req_body = DeleteFeatureTable(name=feature_table, dry_run=dry_run)
        self._call_endpoint(DeleteFeatureTable, req_body, req_context)

    def add_data_sources(
        self,
        feature_table: str,
        tables: List[str],
        paths: List[str],
        custom_sources: List[str],
        req_context: RequestContext,
    ):
        req_body = AddDataSources(
            feature_table=feature_table,
            tables=tables,
            paths=paths,
            custom_sources=custom_sources,
        )
        self._call_endpoint(AddDataSources, req_body, req_context)

    def delete_data_sources(
        self,
        feature_table: str,
        source_names: List[str],
        req_context: RequestContext,
    ):
        req_body = DeleteDataSources(feature_table=feature_table, sources=source_names)
        self._call_endpoint(DeleteDataSources, req_body, req_context)

    def add_notebook_producer(
        self,
        feature_table: str,
        notebook_id: int,
        revision_id: int,
        producer_action: ProducerAction,
        req_context: RequestContext,
    ):
        notebook = Notebook(
            notebook_id=notebook_id,
            revision_id=revision_id,
            notebook_workspace_id=self._local_workspace_id,
            notebook_workspace_url=self._local_host,
        )
        req_body = AddProducer(
            feature_table=feature_table,
            notebook=notebook,
            producer_action=producer_action,
        )
        self._call_endpoint(AddProducer, req_body, req_context)

    def add_job_producer(
        self,
        feature_table: str,
        job_id: int,
        run_id: int,
        producer_action: ProducerAction,
        req_context: RequestContext,
    ):
        job = Job(
            job_id=job_id,
            run_id=run_id,
            job_workspace_id=self._local_workspace_id,
            job_workspace_url=self._local_host,
        )
        req_body = AddProducer(
            feature_table=feature_table, job_run=job, producer_action=producer_action
        )
        self._call_endpoint(AddProducer, req_body, req_context)

    def add_notebook_consumer(
        self,
        feature_table_map: Dict[str, List[str]],
        notebook_id: int,
        revision_id: int,
        req_context: RequestContext,
    ):
        features = [
            ConsumedFeatures(
                table=feature_table,
                names=features,
            )
            for feature_table, features in feature_table_map.items()
        ]
        notebook = Notebook(
            notebook_id=notebook_id,
            revision_id=revision_id,
            notebook_workspace_id=self._local_workspace_id,
            notebook_workspace_url=self._local_host,
        )
        req_body = AddConsumer(
            features=features,
            notebook=notebook,
        )
        self._call_endpoint(AddConsumer, req_body, req_context)

    def add_job_consumer(
        self,
        feature_table_map: Dict[str, List[str]],
        job_id: int,
        run_id: int,
        req_context: RequestContext,
    ):
        features = [
            ConsumedFeatures(
                table=feature_table,
                names=features,
            )
            for feature_table, features in feature_table_map.items()
        ]
        job = Job(
            job_id=job_id,
            run_id=run_id,
            job_workspace_id=self._local_workspace_id,
            job_workspace_url=self._local_host,
        )
        req_body = AddConsumer(
            features=features,
            job_run=job,
        )
        self._call_endpoint(AddConsumer, req_body, req_context)

    def get_consumers(self, feature_table: str, req_context: RequestContext):
        req_body = GetConsumers(feature_table=feature_table)
        response_proto = self._call_endpoint(GetConsumers, req_body, req_context)
        return [Consumer.from_proto(consumer) for consumer in response_proto.consumers]

    def get_model_serving_metadata(
        self,
        feature_table_to_features: Dict[str, List[str]],
        req_context: RequestContext,
    ):
        req_body = GetModelServingMetadata(
            feature_table_features=[
                FeatureTableFeatures(feature_table_name=ft_name, features=features)
                for (ft_name, features) in feature_table_to_features.items()
            ]
        )
        response_proto = self._call_endpoint(
            GetModelServingMetadata, req_body, req_context
        )
        return [
            OnlineFeatureTable.from_proto(online_ft)
            for online_ft in response_proto.online_feature_tables
        ]

    def set_feature_table_tags(
        self, feature_table_id: str, tags: Dict[str, str], req_context: RequestContext
    ) -> None:
        proto_tags = [ProtoTag(key=key, value=value) for key, value in tags.items()]
        req_body = SetTags(feature_table_id=feature_table_id, tags=proto_tags)
        self._call_endpoint(SetTags, req_body, req_context)

    def delete_feature_table_tags(
        self, feature_table_id: str, tags: List[str], req_context: RequestContext
    ) -> None:
        req_body = DeleteTags(feature_table_id=feature_table_id, keys=tags)
        self._call_endpoint(DeleteTags, req_body, req_context)

    def get_feature_table_tags(
        self, feature_table_id: str, req_context: RequestContext
    ) -> List[Tag]:
        req_body = GetTags(feature_table_id=feature_table_id)
        response_proto = self._call_endpoint(GetTags, req_body, req_context)
        return [Tag.from_proto(tag_proto) for tag_proto in response_proto.tags]

    def get_online_store(
        self,
        feature_table: str,
        online_store_metadata: OnlineStoreMetadata,
        req_context: RequestContext,
    ):
        req_body = GetOnlineStore(
            feature_table=feature_table,
            online_table=online_store_metadata.online_table,
            cloud=online_store_metadata.cloud,
            store_type=online_store_metadata.store_type,
        )
        if online_store_metadata.store_type == StoreType.DYNAMODB:
            req_body.table_arn = online_store_metadata.additional_metadata.table_arn
        elif online_store_metadata.store_type == StoreType.COSMOSDB:
            req_body.container_uri = (
                online_store_metadata.additional_metadata.container_uri
            )
        response_proto = self._call_endpoint(GetOnlineStore, req_body, req_context)
        return OnlineStoreDetailed.from_proto(response_proto.online_store)
